# 68 "src/reason-parser/reason_declarative_lexer.mll"
 
open Lexing
open Reason_parser
open Reason_errors

(* The table of keywords *)

let keyword_table, reverse_keyword_table =
  let create_hashtable n l =
    let t = Hashtbl.create n in
    let rev_t = Hashtbl.create n in
    List.iter (fun (k, v) ->
      Hashtbl.add t k v;
      Hashtbl.add rev_t v k;
    ) l;
    t, rev_t
  in
  create_hashtable 149 [
    "and", AND;
    "as", AS;
    "assert", ASSERT;
    "begin", BEGIN;
    "class", CLASS;
    "constraint", CONSTRAINT;
    "do", DO;
    "done", DONE;
    "downto", DOWNTO;
    "else", ELSE;
    "end", END;
    "exception", EXCEPTION;
    "external", EXTERNAL;
    "false", FALSE;
    "for", FOR;
    "fun", FUN;
    "esfun", ES6_FUN;
    "function", FUNCTION;
    "functor", FUNCTOR;
    "if", IF;
    "in", IN;
    "include", INCLUDE;
    "inherit", INHERIT;
    "initializer", INITIALIZER;
    "lazy", LAZY;
    "let", LET;
    "switch", SWITCH;
    "module", MODULE;
    "pub", PUB;
    "mutable", MUTABLE;
    "new", NEW;
    "nonrec", NONREC;
    "object", OBJECT;
    "of", OF;
    "open", OPEN;
    "or", OR;
(*  "parser", PARSER; *)
    "pri", PRI;
    "rec", REC;
    "sig", SIG;
    "struct", STRUCT;
    "then", THEN;
    "to", TO;
    "true", TRUE;
    "try", TRY;
    "type", TYPE;
    "val", VAL;
    "virtual", VIRTUAL;
    "when", WHEN;
    "while", WHILE;
    "with", WITH;

    "mod", INFIXOP3("mod");
    "land", INFIXOP3("land");
    "lor", INFIXOP3("lor");
    "lxor", INFIXOP3("lxor");
    "lsl", INFIXOP4("lsl");
    "lsr", INFIXOP4("lsr");
    "asr", INFIXOP4("asr")
]

(* The only internal state of the lexer is two scratch buffers.
   They could be allocated everytime they are needed, but
   for better performance (FIXME: does this really matter?)
   they are preallocated.*)

type state = {
  raw_buffer : Buffer.t;
  txt_buffer : Buffer.t;
}

let get_scratch_buffers { raw_buffer; txt_buffer } =
  Buffer.reset raw_buffer;
  Buffer.reset txt_buffer;
  ( raw_buffer, txt_buffer )

let flush_buffer buffer =
  let result = Buffer.contents buffer in
  Buffer.reset buffer;
  result

let make () = {
  raw_buffer = Buffer.create 255;
  txt_buffer = Buffer.create 255;
}

(* Specialize raise_error for lexing errors *)

let raise_error loc error = raise_error (Lexing_error error) loc

let store_lexeme buffer lexbuf =
  Buffer.add_string buffer (Lexing.lexeme lexbuf)

(* To "unlex" a few characters *)
let set_lexeme_length buf n = (
  let open Lexing in
  if n < 0 then
    invalid_arg "set_lexeme_length: offset should be positive";
  if n > buf.lex_curr_pos - buf.lex_start_pos then
    invalid_arg "set_lexeme_length: offset larger than lexeme";
  buf.lex_curr_pos <- buf.lex_start_pos + n;
  buf.lex_curr_p <- {buf.lex_start_p
                     with pos_cnum = buf.lex_abs_pos + buf.lex_curr_pos};
)

(* This cut comment characters of the current buffer.
 * Operators (including "/*" and "//") are lexed with the same rule, and this
 * function cuts the lexeme at the beginning of an operator. *)
let lexeme_without_comment buf = (
  let lexeme = Lexing.lexeme buf in
  let i = ref 0 and len = String.length lexeme - 1 in
  let found = ref (-1) in
  while !i < len && !found = -1 do
    begin match lexeme.[!i], lexeme.[!i+1] with
      | ('/', '*') | ('/', '/') | ('*', '/') ->
        found := !i;
      | _ -> ()
    end;
    incr i
  done;
  match !found with
  | -1 -> lexeme
  | n ->
      set_lexeme_length buf n;
      String.sub lexeme 0 n
)

(* Operators that could conflict with comments (those containing /*, */ and //)
 * are escaped in the source. The lexer removes the escapes so that the
 * identifier looks like OCaml ones.
 * An escape in first position is kept to distinguish "verbatim" operators
 * (\=== for instance). *)
let unescape_operator str =
  if (str <> "" && String.contains_from str 1 '\\') then (
    let b = Buffer.create (String.length str) in
    Buffer.add_char b str.[0];
    for i = 1 to String.length str - 1 do
      let c = str.[i] in
      if c <> '\\' then Buffer.add_char b c
    done;
    Buffer.contents b
  ) else str

let lexeme_operator lexbuf =
  unescape_operator (lexeme_without_comment lexbuf)

(* To translate escape sequences *)

let char_for_backslash = function
  | 'n' -> '\010'
  | 'r' -> '\013'
  | 'b' -> '\008'
  | 't' -> '\009'
  | c   -> c

let char_for_decimal_code lexbuf i =
  let c = 100 * (Char.code(Lexing.lexeme_char lexbuf i) - 48) +
           10 * (Char.code(Lexing.lexeme_char lexbuf (i+1)) - 48) +
                (Char.code(Lexing.lexeme_char lexbuf (i+2)) - 48) in
  if (c < 0 || c > 255) then (
    raise_error
        (Location.curr lexbuf)
        (Illegal_escape (Lexing.lexeme lexbuf));
    'x'
  ) else Char.chr c

let char_for_hexadecimal_code lexbuf i =
  let d1 = Char.code (Lexing.lexeme_char lexbuf i) in
  let val1 = if d1 >= 97 then d1 - 87
             else if d1 >= 65 then d1 - 55
             else d1 - 48
  in
  let d2 = Char.code (Lexing.lexeme_char lexbuf (i+1)) in
  let val2 = if d2 >= 97 then d2 - 87
             else if d2 >= 65 then d2 - 55
             else d2 - 48
  in
  Char.chr (val1 * 16 + val2)

(* To convert integer literals, allowing max_int + 1 (PR#4210) *)

let cvt_int_literal s =
  - int_of_string ("-" ^ s)
let cvt_int32_literal s =
  Int32.neg (Int32.of_string ("-" ^ String.sub s 0 (String.length s - 1)))
let cvt_int64_literal s =
  Int64.neg (Int64.of_string ("-" ^ String.sub s 0 (String.length s - 1)))
let cvt_nativeint_literal s =
  Nativeint.neg (Nativeint.of_string ("-" ^ String.sub s 0
                                                       (String.length s - 1)))

(* Remove underscores from float literals *)

let remove_underscores s =
  let l = String.length s in
  let b = Bytes.create l in
  let rec remove src dst =
    if src >= l then
      if dst >= l then s else Bytes.sub_string b 0 dst
    else
      match s.[src] with
        '_' -> remove (src + 1) dst
      |  c  -> Bytes.set b dst c; remove (src + 1) (dst + 1)
  in remove 0 0

(* Update the current location with file name and line number. *)

let update_loc lexbuf file line absolute chars =
  let pos = lexbuf.lex_curr_p in
  let new_file = match file with
                 | None -> pos.pos_fname
                 | Some s -> s
  in
  lexbuf.lex_curr_p <- { pos with
    pos_fname = new_file;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }


# 241 "src/reason-parser/reason_declarative_lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\146\255\147\255\224\000\003\001\038\001\073\001\108\001\
    \143\001\178\001\184\255\213\001\191\255\250\001\091\000\063\002\
    \068\000\071\000\084\000\155\002\219\255\190\002\221\255\222\255\
    \224\255\225\002\064\003\012\001\094\003\238\255\178\003\006\004\
    \090\004\042\005\250\005\202\006\154\007\106\008\073\009\108\009\
    \143\009\173\009\122\000\254\255\001\000\163\010\198\010\233\010\
    \012\011\047\011\082\011\255\255\005\000\117\011\093\000\152\011\
    \187\011\094\000\222\011\095\000\096\000\097\000\006\000\103\000\
    \033\012\125\012\160\012\227\012\055\013\148\004\139\013\106\014\
    \073\015\157\015\124\016\091\017\175\017\003\018\087\018\171\018\
    \255\018\083\019\167\019\251\019\079\020\163\020\247\020\108\000\
    \075\021\159\021\243\021\071\022\155\022\239\022\067\023\151\023\
    \235\023\118\000\063\024\103\005\237\255\062\003\021\000\177\024\
    \005\025\196\255\011\001\193\255\096\025\108\000\109\000\007\000\
    \236\255\235\255\231\255\242\002\128\000\110\000\234\255\011\002\
    \113\000\233\255\069\006\147\000\232\255\060\007\124\006\121\025\
    \216\000\162\025\197\025\227\255\011\000\012\000\008\001\024\001\
    \016\000\227\255\017\000\232\025\011\026\046\026\141\000\212\255\
    \208\255\209\255\210\255\206\255\081\026\116\026\183\026\011\027\
    \095\027\179\027\220\007\007\028\091\028\181\028\244\028\102\029\
    \192\255\015\001\102\029\194\029\151\000\176\255\200\255\201\255\
    \129\000\180\255\174\255\005\030\183\255\097\030\132\030\167\030\
    \164\008\234\030\231\009\186\255\070\031\105\031\140\031\175\031\
    \181\255\242\031\197\013\078\032\113\032\148\032\183\032\218\032\
    \010\001\012\001\253\032\040\001\046\001\032\033\065\001\075\001\
    \248\255\143\000\079\001\081\001\254\000\254\255\255\255\250\255\
    \100\001\251\255\149\000\249\255\089\007\246\255\247\255\018\000\
    \248\255\064\002\134\014\253\255\200\000\223\000\255\255\254\255\
    \252\255\162\014\079\033\234\000\236\000\139\001\251\255\250\255\
    \249\255\240\011\155\002\252\000\081\003\005\001\178\012\014\001\
    \146\001\199\001\247\255\248\255\249\255\147\001\112\033\255\255\
    \250\255\232\015\091\003\253\255\046\001\148\001\164\001\223\004\
    \252\255\201\016\251\255\174\001\234\001\252\255\136\033\254\255\
    \255\255\181\001\182\001\253\255\165\033\033\001\037\001\069\001\
    \079\001\039\001\080\001\056\001\013\000\255\255";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\103\000\098\000\097\000\096\000\087\000\
    \082\000\104\000\255\255\067\000\255\255\070\000\052\000\050\000\
    \048\000\044\000\041\000\090\000\255\255\035\000\255\255\255\255\
    \255\255\029\000\040\000\032\000\065\000\255\255\011\000\011\000\
    \010\000\009\000\008\000\007\000\007\000\007\000\051\000\005\000\
    \004\000\003\000\002\000\255\255\109\000\109\000\104\000\104\000\
    \100\000\255\255\255\255\255\255\255\255\095\000\255\255\085\000\
    \086\000\255\255\102\000\255\255\255\255\255\255\255\255\255\255\
    \096\000\038\000\006\000\096\000\039\000\073\000\007\000\007\000\
    \106\000\007\000\007\000\107\000\015\000\015\000\013\000\012\000\
    \015\000\012\000\012\000\011\000\013\000\012\000\013\000\255\255\
    \013\000\015\000\014\000\014\000\014\000\011\000\011\000\015\000\
    \013\000\255\255\013\000\066\000\255\255\255\255\061\000\060\000\
    \060\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\024\000\024\000\024\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\027\000\
    \255\255\026\000\025\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\030\000\091\000\037\000\042\000\255\255\
    \255\255\255\255\255\255\255\255\096\000\092\000\096\000\255\255\
    \056\000\057\000\255\255\255\255\056\000\093\000\255\255\255\255\
    \255\255\255\255\096\000\094\000\080\000\255\255\255\255\255\255\
    \053\000\255\255\255\255\096\000\255\255\096\000\096\000\058\000\
    \078\000\096\000\077\000\255\255\068\000\101\000\104\000\104\000\
    \255\255\104\000\076\000\083\000\084\000\099\000\089\000\088\000\
    \255\255\255\255\098\000\255\255\255\255\105\000\255\255\255\255\
    \255\255\007\000\007\000\002\000\255\255\255\255\255\255\255\255\
    \003\000\255\255\002\000\255\255\255\255\255\255\255\255\009\000\
    \255\255\009\000\009\000\255\255\009\000\009\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\008\000\008\000\255\255\
    \255\255\005\000\005\000\255\255\001\000\005\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\003\000\255\255\
    \255\255\003\000\255\255\255\255\255\255\002\000\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\000\000\000\000\
    \000\000\255\255\255\255\109\000\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\000\000\255\255\000\000\114\000\255\255\255\255\255\255\
    \000\000\000\000\000\000\255\255\255\255\255\255\000\000\255\255\
    \255\255\000\000\255\255\255\255\000\000\133\000\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\133\000\134\000\133\000\
    \136\000\000\000\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\000\000\000\000\000\000\
    \255\255\000\000\000\000\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\200\000\
    \000\000\255\255\255\255\255\255\204\000\000\000\000\000\000\000\
    \255\255\000\000\255\255\000\000\213\000\000\000\000\000\255\255\
    \000\000\227\000\255\255\000\000\255\255\255\255\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\242\000\000\000\000\000\000\000\255\255\248\000\000\000\
    \000\000\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \000\000\255\255\000\000\255\255\005\001\000\000\255\255\000\000\
    \000\000\255\255\255\255\000\000\255\255\255\255\255\255\015\001\
    \018\001\255\255\018\001\255\255\255\255\000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\042\000\043\000\043\000\042\000\044\000\062\000\051\000\
    \043\000\110\000\052\000\062\000\111\000\131\000\131\000\021\001\
    \132\000\132\000\137\000\137\000\214\000\138\000\138\000\240\000\
    \042\000\008\000\029\000\026\000\006\000\003\000\025\000\027\000\
    \023\000\022\000\021\000\007\000\020\000\019\000\018\000\009\000\
    \031\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\017\000\016\000\015\000\038\000\013\000\039\000\
    \005\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\014\000\045\000\012\000\004\000\041\000\
    \024\000\036\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\037\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\028\000\011\000\010\000\040\000\147\000\
    \164\000\146\000\142\000\042\000\145\000\144\000\042\000\053\000\
    \048\000\058\000\058\000\046\000\053\000\048\000\046\000\058\000\
    \046\000\006\000\107\000\113\000\112\000\118\000\006\000\167\000\
    \121\000\166\000\042\000\165\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\124\000\143\000\170\000\169\000\211\000\210\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\168\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\223\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \002\000\197\000\127\000\197\000\197\000\197\000\197\000\127\000\
    \206\000\222\000\197\000\197\000\255\255\197\000\197\000\197\000\
    \255\255\232\000\131\000\231\000\106\000\132\000\110\000\106\000\
    \161\000\111\000\197\000\161\000\197\000\197\000\197\000\197\000\
    \197\000\135\000\131\000\232\000\194\000\132\000\194\000\194\000\
    \194\000\194\000\136\000\106\000\232\000\194\000\194\000\161\000\
    \194\000\194\000\194\000\255\255\189\000\232\000\005\000\252\000\
    \135\000\189\000\134\000\005\000\198\000\194\000\197\000\194\000\
    \194\000\194\000\194\000\194\000\014\001\105\000\015\001\005\000\
    \019\001\005\000\005\000\005\000\005\000\160\000\252\000\016\001\
    \005\000\005\000\194\000\005\000\005\000\005\000\194\000\194\000\
    \194\000\017\001\017\001\020\001\197\000\194\000\197\000\195\000\
    \005\000\194\000\005\000\005\000\005\000\005\000\005\000\000\000\
    \108\000\000\000\006\000\197\000\006\000\006\000\006\000\006\000\
    \197\000\000\000\000\000\006\000\006\000\201\000\006\000\006\000\
    \006\000\203\000\202\000\208\000\000\000\000\000\204\000\194\000\
    \207\000\194\000\193\000\006\000\005\000\006\000\006\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\048\000\210\000\048\000\
    \048\000\048\000\048\000\209\000\000\000\228\000\048\000\189\000\
    \229\000\048\000\191\000\048\000\214\000\244\000\252\000\240\000\
    \003\001\254\000\005\000\000\000\005\000\063\000\048\000\006\000\
    \048\000\190\000\048\000\048\000\048\000\000\000\252\000\000\000\
    \053\000\254\000\053\000\053\000\053\000\053\000\000\000\000\000\
    \244\000\053\000\053\000\003\001\053\000\053\000\053\000\008\001\
    \008\001\000\000\010\001\010\001\000\000\006\000\000\000\006\000\
    \057\000\053\000\048\000\053\000\187\000\053\000\053\000\053\000\
    \000\000\244\000\000\000\046\000\245\000\046\000\046\000\046\000\
    \046\000\000\000\000\000\000\000\046\000\046\000\000\000\046\000\
    \046\000\046\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \048\000\247\000\048\000\054\000\046\000\053\000\046\000\046\000\
    \181\000\046\000\046\000\000\000\008\001\000\000\006\000\009\001\
    \006\000\006\000\006\000\006\000\000\000\000\000\205\000\006\000\
    \006\000\000\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\053\000\255\255\053\000\061\000\006\000\
    \046\000\006\000\006\000\006\000\006\000\006\000\000\000\000\000\
    \255\255\000\000\000\000\006\000\000\000\006\000\006\000\006\000\
    \006\000\000\000\000\000\246\000\006\000\006\000\000\000\006\000\
    \173\000\006\000\000\000\000\000\000\000\000\000\046\000\000\000\
    \046\000\063\000\179\000\006\000\006\000\000\000\171\000\006\000\
    \006\000\006\000\006\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\120\000\120\000\255\255\000\000\000\000\
    \000\000\000\000\228\000\255\255\000\000\229\000\000\000\255\255\
    \255\255\180\000\000\000\006\000\000\000\000\000\063\000\000\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \006\000\000\000\006\000\006\000\006\000\006\000\006\001\230\000\
    \000\000\006\000\006\000\000\000\006\000\148\000\150\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\006\000\172\000\
    \006\000\006\000\000\000\006\000\006\000\149\000\006\000\006\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\000\000\063\000\226\000\006\000\151\000\000\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\000\000\006\000\048\000\006\000\048\000\048\000\
    \048\000\048\000\000\000\000\000\000\000\048\000\048\000\243\000\
    \048\000\140\000\048\000\236\000\236\000\236\000\236\000\236\000\
    \236\000\236\000\236\000\236\000\236\000\048\000\000\000\048\000\
    \048\000\141\000\048\000\048\000\000\000\000\000\000\000\046\000\
    \000\000\046\000\046\000\046\000\046\000\000\000\000\000\000\000\
    \058\000\046\000\007\001\046\000\046\000\046\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\057\000\
    \046\000\048\000\046\000\046\000\046\000\046\000\046\000\000\000\
    \000\000\000\000\006\000\000\000\006\000\006\000\006\000\139\000\
    \000\000\000\000\000\000\006\000\006\000\000\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\000\000\048\000\
    \000\000\048\000\059\000\006\000\046\000\006\000\006\000\006\000\
    \006\000\006\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\000\000\046\000\000\000\046\000\063\000\000\000\006\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\126\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\000\000\000\000\000\000\000\000\006\000\000\000\006\000\
    \126\000\127\000\000\000\127\000\127\000\127\000\127\000\000\000\
    \000\000\000\000\127\000\127\000\000\000\127\000\127\000\127\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\127\000\000\000\127\000\129\000\127\000\127\000\
    \127\000\237\000\237\000\237\000\237\000\237\000\237\000\237\000\
    \237\000\237\000\237\000\255\000\255\000\255\000\255\000\255\000\
    \255\000\255\000\255\000\255\000\255\000\000\000\000\000\000\000\
    \000\000\000\000\099\000\000\000\128\000\101\000\127\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\000\000\100\000\000\000\127\000\101\000\127\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\076\000\100\000\000\000\000\000\000\000\000\000\000\000\
    \078\000\000\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\076\000\076\000\076\000\076\000\077\000\
    \076\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\000\000\000\000\000\000\
    \000\000\030\000\000\000\076\000\076\000\076\000\076\000\077\000\
    \076\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\076\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\078\000\000\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\076\000\
    \080\000\076\000\076\000\077\000\076\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\081\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\082\000\079\000\
    \079\000\000\000\000\000\000\000\000\000\030\000\000\000\076\000\
    \080\000\076\000\076\000\077\000\076\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\081\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\082\000\079\000\
    \079\000\032\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\000\000\000\000\000\000\
    \000\000\032\000\000\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\000\000\
    \000\000\000\000\000\000\069\000\000\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\000\001\
    \000\001\000\001\000\001\000\001\000\001\000\001\000\001\000\001\
    \000\001\000\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\000\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\033\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\000\000\000\000\000\000\
    \000\000\033\000\000\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\102\000\000\000\000\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\000\000\000\000\000\000\000\000\103\000\000\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\000\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\034\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\000\000\000\000\000\000\
    \000\000\034\000\000\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\126\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\126\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\000\000\000\000\
    \000\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\035\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\000\000\000\000\000\000\
    \000\000\035\000\000\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\135\000\131\000\000\000\
    \000\000\132\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\135\000\000\000\134\000\000\000\
    \000\000\000\000\000\000\214\000\000\000\000\000\215\000\000\000\
    \000\000\000\000\000\000\000\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\000\000\000\000\
    \000\000\000\000\000\000\219\000\000\000\000\000\000\000\000\000\
    \217\000\000\000\000\000\220\000\000\000\000\000\000\000\000\000\
    \221\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\035\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\218\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\000\000\000\000\000\000\
    \000\000\035\000\000\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \073\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\000\000\
    \000\000\000\000\000\000\151\000\255\255\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\000\000\
    \000\000\216\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\035\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\000\000\000\000\000\000\
    \000\000\035\000\000\000\035\000\035\000\035\000\035\000\070\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\000\000\
    \000\000\000\000\000\000\176\000\000\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\006\000\000\000\006\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\006\000\006\000\000\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\006\000\000\000\064\000\006\000\065\000\
    \066\000\006\000\000\000\000\000\000\000\053\000\000\000\053\000\
    \053\000\053\000\053\000\000\000\000\000\000\000\053\000\053\000\
    \000\000\053\000\053\000\053\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\063\000\053\000\006\000\
    \053\000\053\000\053\000\053\000\053\000\000\000\000\000\000\000\
    \053\000\000\000\053\000\053\000\053\000\053\000\000\000\000\000\
    \000\000\053\000\053\000\000\000\053\000\053\000\053\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\006\000\000\000\006\000\
    \054\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \000\000\000\000\000\000\000\000\035\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\000\000\
    \053\000\000\000\053\000\054\000\000\000\053\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \000\000\000\000\000\000\053\000\035\000\053\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\000\000\000\000\000\000\000\000\178\000\000\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\000\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\000\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\051\000\000\000\000\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\050\000\000\000\000\000\006\000\
    \000\000\006\000\000\000\000\000\000\000\047\000\048\000\000\000\
    \048\000\000\000\046\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\006\000\
    \006\000\006\000\049\000\005\000\000\000\000\000\000\000\046\000\
    \000\000\046\000\046\000\046\000\046\000\000\000\000\000\000\000\
    \046\000\046\000\000\000\046\000\046\000\046\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \046\000\004\000\046\000\046\000\046\000\046\000\046\000\000\000\
    \000\000\000\000\046\000\000\000\046\000\046\000\046\000\046\000\
    \000\000\000\000\000\000\058\000\046\000\000\000\046\000\046\000\
    \046\000\000\000\000\000\000\000\000\000\000\000\000\000\006\000\
    \000\000\049\000\061\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\000\000\000\000\000\000\048\000\000\000\048\000\
    \048\000\048\000\048\000\000\000\000\000\000\000\048\000\048\000\
    \000\000\048\000\048\000\048\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\046\000\000\000\046\000\059\000\048\000\046\000\
    \048\000\048\000\048\000\048\000\048\000\000\000\000\000\000\000\
    \053\000\000\000\053\000\053\000\053\000\053\000\000\000\000\000\
    \000\000\053\000\053\000\000\000\053\000\053\000\053\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\046\000\000\000\046\000\
    \057\000\053\000\048\000\053\000\053\000\053\000\053\000\053\000\
    \000\000\000\000\000\000\053\000\000\000\053\000\053\000\053\000\
    \053\000\000\000\000\000\000\000\053\000\053\000\000\000\053\000\
    \053\000\053\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \048\000\000\000\048\000\054\000\053\000\053\000\053\000\055\000\
    \053\000\053\000\053\000\000\000\000\000\000\000\053\000\000\000\
    \053\000\053\000\053\000\053\000\000\000\000\000\000\000\053\000\
    \053\000\000\000\053\000\053\000\053\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\053\000\000\000\053\000\054\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\000\000\000\000\
    \000\000\053\000\000\000\053\000\053\000\053\000\053\000\000\000\
    \000\000\000\000\053\000\053\000\000\000\053\000\053\000\053\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\053\000\000\000\
    \053\000\054\000\053\000\053\000\053\000\056\000\053\000\053\000\
    \053\000\000\000\000\000\000\000\053\000\000\000\053\000\053\000\
    \053\000\053\000\000\000\000\000\000\000\053\000\053\000\000\000\
    \053\000\053\000\053\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\053\000\000\000\053\000\054\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\000\000\000\000\000\000\058\000\
    \000\000\058\000\058\000\058\000\058\000\000\000\000\000\000\000\
    \058\000\058\000\000\000\058\000\058\000\058\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\053\000\000\000\053\000\054\000\
    \058\000\053\000\058\000\058\000\058\000\058\000\058\000\000\000\
    \238\000\238\000\238\000\238\000\238\000\238\000\238\000\238\000\
    \238\000\238\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\238\000\238\000\238\000\238\000\238\000\238\000\053\000\
    \000\000\053\000\060\000\000\000\058\000\000\000\000\000\000\000\
    \000\000\000\000\006\000\000\000\006\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\006\000\006\000\000\000\006\000\006\000\
    \006\000\238\000\238\000\238\000\238\000\238\000\238\000\000\000\
    \000\000\000\000\058\000\006\000\058\000\006\000\006\000\006\000\
    \006\000\006\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\000\000\063\000\000\000\006\000\
    \069\000\000\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\000\000\006\000\006\000\006\000\
    \006\000\006\000\006\000\006\000\000\000\000\000\000\000\006\000\
    \006\000\000\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\006\000\
    \000\000\067\000\006\000\006\000\006\000\006\000\000\000\000\000\
    \000\000\006\000\000\000\006\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\006\000\006\000\000\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\063\000\006\000\006\000\006\000\006\000\006\000\006\000\
    \006\000\000\000\239\000\239\000\239\000\239\000\239\000\239\000\
    \239\000\239\000\239\000\239\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\239\000\239\000\239\000\239\000\239\000\
    \239\000\006\000\000\000\006\000\063\000\000\000\006\000\000\000\
    \000\000\000\000\000\000\000\000\006\000\000\000\006\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\006\000\006\000\000\000\
    \006\000\006\000\006\000\239\000\239\000\239\000\239\000\239\000\
    \239\000\000\000\000\000\000\000\006\000\006\000\006\000\006\000\
    \006\000\006\000\006\000\006\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\000\000\063\000\
    \000\000\006\000\068\000\000\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\006\000\
    \000\000\006\000\000\000\000\000\000\000\068\000\000\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\000\000\000\000\000\000\000\000\068\000\000\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\035\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\000\000\000\000\
    \000\000\000\000\035\000\000\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\071\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \000\000\000\000\000\000\000\000\186\000\000\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\000\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\000\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\072\000\000\000\000\000\072\000\000\000\
    \072\000\035\000\000\000\000\000\072\000\072\000\000\000\072\000\
    \072\000\072\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\000\000\000\000\072\000\072\000\
    \072\000\000\000\072\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\000\000\000\000\000\000\
    \072\000\035\000\000\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\225\000\072\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\224\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\000\000\224\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\000\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\072\000\000\000\000\000\072\000\072\000\072\000\
    \000\000\000\000\000\000\072\000\072\000\000\000\072\000\000\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\000\000\000\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\000\000\072\000\000\000\072\000\
    \072\000\000\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\035\000\072\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \000\000\000\000\000\000\000\000\035\000\000\000\035\000\035\000\
    \035\000\074\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\
    \001\001\001\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\001\001\001\001\001\001\001\001\001\001\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\001\001\001\001\001\001\001\001\001\001\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\000\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\000\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\075\000\000\000\000\000\
    \075\000\000\000\075\000\035\000\000\000\000\000\075\000\075\000\
    \000\000\075\000\075\000\075\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\000\000\000\000\
    \075\000\075\000\075\000\000\000\075\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\000\000\
    \000\000\000\000\075\000\035\000\000\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\000\000\
    \075\000\002\001\002\001\002\001\002\001\002\001\002\001\002\001\
    \002\001\002\001\002\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\002\001\002\001\002\001\002\001\002\001\002\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\002\001\002\001\002\001\002\001\002\001\002\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\000\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\000\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\075\000\000\000\000\000\075\000\
    \075\000\075\000\000\000\000\000\000\000\075\000\075\000\000\000\
    \075\000\000\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\000\000\000\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\000\000\075\000\
    \000\000\075\000\075\000\000\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\076\000\075\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\000\000\000\000\000\000\000\000\076\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\000\000\000\000\000\000\097\000\000\000\
    \097\000\000\000\000\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\098\000\098\000\098\000\098\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\000\000\000\000\
    \000\000\000\000\076\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\089\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \089\000\089\000\089\000\089\000\095\000\089\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\000\000\000\000\000\000\000\000\078\000\000\000\
    \089\000\089\000\089\000\089\000\095\000\089\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\076\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\000\000\000\000\
    \000\000\000\000\076\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\094\000\
    \094\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\000\000\000\000\000\000\000\000\076\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\093\000\093\000\076\000\076\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\000\000\000\000\
    \000\000\000\000\076\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\000\000\000\000\000\000\000\000\076\000\000\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\084\000\000\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\085\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\000\000\000\000\
    \000\000\000\000\083\000\000\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\085\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\089\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\092\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\000\000\000\000\000\000\000\000\084\000\000\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\092\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\076\000\000\000\000\000\000\000\087\000\000\000\
    \087\000\000\000\000\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\000\000\000\000\
    \000\000\000\000\076\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\000\000\000\000\000\000\000\000\086\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\089\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\000\000\000\000\
    \000\000\000\000\088\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\089\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\000\000\000\000\000\000\000\000\089\000\000\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\000\000\000\000\
    \000\000\000\000\089\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\076\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\000\000\000\000\000\000\000\000\076\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\089\000\000\000\000\000\000\000\087\000\000\000\
    \087\000\000\000\000\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\000\000\000\000\
    \000\000\000\000\089\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\076\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\093\000\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\076\000\
    \076\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\000\000\000\000\000\000\000\000\093\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\076\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\094\000\094\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\000\000\000\000\
    \000\000\000\000\094\000\000\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\089\000\000\000\
    \000\000\000\000\097\000\000\000\097\000\000\000\000\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\000\000\000\000\000\000\000\000\089\000\000\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\000\000\000\000\
    \000\000\000\000\096\000\000\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\076\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\098\000\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\000\000\000\000\000\000\000\000\098\000\000\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\106\000\000\000\000\000\106\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\106\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \103\000\000\000\000\000\000\000\000\000\000\000\000\000\104\000\
    \000\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\105\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\000\000\000\000\000\000\000\000\
    \103\000\000\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\104\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\104\000\000\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \000\000\000\000\000\000\000\000\104\000\000\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \117\000\000\000\117\000\000\000\000\000\000\000\000\000\117\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \116\000\116\000\127\000\000\000\127\000\127\000\127\000\127\000\
    \000\000\000\000\000\000\127\000\127\000\000\000\127\000\127\000\
    \127\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\127\000\000\000\127\000\127\000\127\000\
    \127\000\127\000\000\000\000\000\117\000\000\000\000\000\000\000\
    \000\000\000\000\117\000\127\000\000\000\127\000\127\000\127\000\
    \127\000\000\000\000\000\000\000\127\000\127\000\117\000\127\000\
    \127\000\127\000\117\000\000\000\117\000\128\000\000\000\127\000\
    \115\000\000\000\000\000\000\000\127\000\000\000\130\000\127\000\
    \127\000\127\000\127\000\000\000\000\000\000\000\127\000\000\000\
    \127\000\127\000\127\000\127\000\000\000\000\000\000\000\127\000\
    \127\000\000\000\127\000\127\000\127\000\127\000\000\000\127\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\128\000\127\000\
    \127\000\127\000\127\000\127\000\127\000\127\000\000\000\000\000\
    \000\000\006\000\000\000\006\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\006\000\006\000\000\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\127\000\000\000\
    \127\000\128\000\006\000\127\000\006\000\006\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\048\000\000\000\048\000\048\000\
    \048\000\048\000\000\000\000\000\000\000\048\000\048\000\000\000\
    \048\000\048\000\048\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\127\000\000\000\127\000\063\000\048\000\006\000\048\000\
    \048\000\048\000\048\000\048\000\000\000\000\000\000\000\048\000\
    \000\000\048\000\048\000\048\000\048\000\000\000\000\000\000\000\
    \048\000\048\000\000\000\048\000\048\000\048\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\006\000\000\000\006\000\057\000\
    \048\000\048\000\048\000\048\000\048\000\048\000\048\000\000\000\
    \000\000\000\000\006\000\000\000\006\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\006\000\006\000\000\000\006\000\162\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\000\000\048\000\
    \000\000\048\000\057\000\006\000\048\000\006\000\006\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\006\000\000\000\006\000\
    \006\000\006\000\006\000\000\000\000\000\000\000\006\000\006\000\
    \000\000\006\000\006\000\006\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\048\000\000\000\048\000\063\000\006\000\006\000\
    \006\000\006\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \159\000\000\000\000\000\159\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\006\000\000\000\006\000\
    \063\000\000\000\006\000\000\000\000\000\000\000\000\000\159\000\
    \006\000\000\000\006\000\006\000\006\000\006\000\000\000\000\000\
    \000\000\006\000\006\000\000\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \006\000\006\000\006\000\006\000\006\000\157\000\006\000\006\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\000\000\063\000\000\000\006\000\158\000\000\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\156\000\006\000\000\000\006\000\000\000\000\000\
    \000\000\000\000\000\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\000\000\000\000\
    \000\000\000\000\156\000\000\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\000\000\000\000\000\000\000\000\156\000\000\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\153\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\154\000\000\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\000\000\000\000\
    \000\000\000\000\153\000\000\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\155\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\154\000\000\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\000\000\000\000\000\000\000\000\155\000\000\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\156\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\000\000\000\000\
    \000\000\000\000\156\000\000\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\006\000\000\000\
    \006\000\006\000\006\000\006\000\000\000\000\000\000\000\006\000\
    \006\000\000\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\006\000\
    \000\000\006\000\006\000\006\000\006\000\006\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\161\000\000\000\000\000\
    \161\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\063\000\000\000\006\000\161\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\158\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\158\000\000\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\000\000\000\000\
    \000\000\006\000\160\000\006\000\000\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\000\000\
    \000\000\000\000\000\000\158\000\000\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\159\000\
    \000\000\000\000\159\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\159\000\006\000\
    \000\000\006\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \006\000\006\000\000\000\006\000\006\000\006\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \006\000\000\000\006\000\006\000\163\000\006\000\006\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\000\000\063\000\000\000\006\000\158\000\000\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\000\000\006\000\006\000\006\000\006\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\006\000\006\000\000\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\006\000\000\000\006\000\006\000\
    \006\000\006\000\006\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\063\000\000\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\006\000\000\000\
    \006\000\006\000\006\000\006\000\000\000\000\000\000\000\006\000\
    \006\000\000\000\006\000\006\000\177\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\006\000\006\000\
    \006\000\006\000\006\000\006\000\006\000\006\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \000\000\063\000\000\000\006\000\176\000\000\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \000\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\006\000\006\000\000\000\006\000\174\000\
    \006\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\006\000\000\000\006\000\006\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\006\000\000\000\006\000\
    \006\000\006\000\006\000\000\000\000\000\000\000\006\000\006\000\
    \000\000\006\000\175\000\006\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\063\000\006\000\006\000\
    \006\000\006\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \006\000\000\000\006\000\006\000\006\000\006\000\000\000\000\000\
    \000\000\006\000\006\000\000\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\006\000\000\000\006\000\
    \063\000\006\000\006\000\006\000\006\000\006\000\006\000\006\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \006\000\000\000\006\000\063\000\000\000\006\000\000\000\000\000\
    \000\000\000\000\000\000\006\000\000\000\006\000\006\000\006\000\
    \006\000\000\000\000\000\000\000\006\000\006\000\000\000\006\000\
    \006\000\006\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\006\000\006\000\006\000\006\000\006\000\
    \006\000\006\000\006\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\000\000\063\000\000\000\
    \006\000\178\000\000\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\000\000\006\000\006\000\
    \006\000\006\000\006\000\006\000\006\000\000\000\000\000\000\000\
    \006\000\006\000\000\000\006\000\006\000\006\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \006\000\000\000\006\000\006\000\006\000\006\000\006\000\000\000\
    \000\000\000\000\046\000\000\000\046\000\046\000\046\000\046\000\
    \000\000\000\000\000\000\046\000\046\000\000\000\046\000\046\000\
    \046\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\063\000\046\000\006\000\182\000\046\000\046\000\
    \046\000\046\000\000\000\000\000\000\000\046\000\000\000\046\000\
    \046\000\046\000\046\000\000\000\000\000\000\000\046\000\046\000\
    \000\000\046\000\046\000\185\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\006\000\000\000\006\000\061\000\046\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\000\000\000\000\000\000\
    \046\000\000\000\046\000\046\000\046\000\046\000\000\000\000\000\
    \000\000\046\000\046\000\000\000\046\000\046\000\046\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\183\000\000\000\046\000\
    \061\000\046\000\046\000\046\000\046\000\046\000\046\000\046\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \046\000\000\000\046\000\061\000\184\000\046\000\000\000\000\000\
    \000\000\000\000\000\000\046\000\000\000\046\000\046\000\046\000\
    \046\000\000\000\000\000\000\000\046\000\046\000\000\000\046\000\
    \046\000\046\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\046\000\046\000\046\000\046\000\046\000\
    \046\000\046\000\046\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\000\000\061\000\000\000\
    \046\000\186\000\000\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\000\000\046\000\053\000\
    \046\000\053\000\053\000\053\000\053\000\000\000\000\000\000\000\
    \053\000\053\000\000\000\053\000\053\000\053\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \053\000\000\000\053\000\188\000\053\000\053\000\053\000\000\000\
    \000\000\000\000\053\000\000\000\053\000\053\000\053\000\053\000\
    \000\000\000\000\000\000\053\000\053\000\000\000\053\000\053\000\
    \053\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\054\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\000\000\000\000\000\000\189\000\000\000\189\000\
    \189\000\189\000\189\000\000\000\000\000\000\000\189\000\189\000\
    \000\000\189\000\189\000\189\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\053\000\000\000\053\000\054\000\189\000\053\000\
    \189\000\189\000\189\000\189\000\189\000\000\000\000\000\000\000\
    \048\000\000\000\048\000\048\000\048\000\048\000\000\000\000\000\
    \000\000\048\000\048\000\000\000\048\000\048\000\048\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\053\000\000\000\053\000\
    \192\000\048\000\189\000\048\000\048\000\048\000\048\000\048\000\
    \000\000\000\000\000\000\048\000\000\000\048\000\048\000\048\000\
    \048\000\000\000\000\000\000\000\048\000\048\000\000\000\048\000\
    \048\000\048\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \189\000\000\000\189\000\057\000\048\000\048\000\048\000\048\000\
    \048\000\048\000\048\000\000\000\000\000\000\000\194\000\000\000\
    \194\000\194\000\194\000\194\000\000\000\000\000\000\000\194\000\
    \194\000\000\000\194\000\194\000\194\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\048\000\000\000\048\000\057\000\194\000\
    \048\000\194\000\194\000\194\000\194\000\194\000\000\000\000\000\
    \000\000\197\000\000\000\197\000\197\000\197\000\197\000\000\000\
    \000\000\000\000\197\000\197\000\000\000\197\000\197\000\197\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\048\000\000\000\
    \048\000\196\000\197\000\194\000\197\000\197\000\197\000\197\000\
    \197\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\235\000\
    \000\000\235\000\000\000\000\000\000\000\000\000\235\000\000\000\
    \000\000\194\000\252\000\194\000\198\000\253\000\197\000\234\000\
    \234\000\234\000\234\000\234\000\234\000\234\000\234\000\234\000\
    \234\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \251\000\000\000\251\000\000\000\000\000\000\000\000\000\251\000\
    \000\000\000\000\000\000\000\000\197\000\000\000\197\000\000\000\
    \250\000\250\000\250\000\250\000\250\000\250\000\250\000\250\000\
    \250\000\250\000\000\000\235\000\000\000\000\000\000\000\000\000\
    \000\000\235\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\235\000\000\000\000\000\
    \000\000\235\000\000\000\235\000\000\000\000\000\000\000\233\000\
    \000\000\000\000\000\000\000\000\251\000\000\000\000\000\000\000\
    \000\000\000\000\251\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\251\000\000\000\
    \000\000\000\000\251\000\000\000\251\000\000\000\000\000\012\001\
    \249\000\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \012\001\012\001\012\001\000\000\012\001\011\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \000\000\000\000\011\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\044\000\000\000\000\000\044\000\052\000\
    \062\000\111\000\052\000\062\000\111\000\132\000\133\000\020\001\
    \132\000\133\000\136\000\138\000\215\000\136\000\138\000\215\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\016\000\
    \014\000\017\000\018\000\042\000\017\000\017\000\042\000\054\000\
    \057\000\059\000\060\000\061\000\054\000\057\000\059\000\060\000\
    \061\000\063\000\102\000\109\000\110\000\117\000\063\000\014\000\
    \120\000\014\000\042\000\014\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\097\000\097\000\
    \097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\116\000\116\000\
    \116\000\116\000\123\000\142\000\164\000\168\000\201\000\210\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\014\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\220\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\128\000\003\000\003\000\003\000\003\000\128\000\
    \204\000\221\000\003\000\003\000\133\000\003\000\003\000\003\000\
    \136\000\227\000\134\000\228\000\106\000\134\000\027\000\106\000\
    \161\000\027\000\003\000\161\000\003\000\003\000\003\000\003\000\
    \003\000\135\000\135\000\235\000\004\000\135\000\004\000\004\000\
    \004\000\004\000\134\000\106\000\237\000\004\000\004\000\161\000\
    \004\000\004\000\004\000\027\000\192\000\239\000\193\000\252\000\
    \135\000\192\000\135\000\193\000\003\000\004\000\003\000\004\000\
    \004\000\004\000\004\000\004\000\013\001\106\000\014\001\005\000\
    \017\001\005\000\005\000\005\000\005\000\161\000\252\000\015\001\
    \005\000\005\000\195\000\005\000\005\000\005\000\195\000\195\000\
    \196\000\016\001\018\001\019\001\003\000\196\000\003\000\004\000\
    \005\000\004\000\005\000\005\000\005\000\005\000\005\000\255\255\
    \027\000\255\255\006\000\198\000\006\000\006\000\006\000\006\000\
    \198\000\255\255\255\255\006\000\006\000\199\000\006\000\006\000\
    \006\000\202\000\199\000\203\000\255\255\255\255\202\000\004\000\
    \203\000\004\000\005\000\006\000\005\000\006\000\006\000\006\000\
    \006\000\006\000\255\255\255\255\255\255\007\000\208\000\007\000\
    \007\000\007\000\007\000\208\000\255\255\229\000\007\000\007\000\
    \229\000\007\000\007\000\007\000\240\000\245\000\253\000\240\000\
    \245\000\253\000\005\000\255\255\005\000\006\000\007\000\006\000\
    \007\000\007\000\007\000\007\000\007\000\255\255\254\000\255\255\
    \008\000\254\000\008\000\008\000\008\000\008\000\255\255\255\255\
    \003\001\008\000\008\000\003\001\008\000\008\000\008\000\009\001\
    \010\001\255\255\009\001\010\001\255\255\006\000\255\255\006\000\
    \007\000\008\000\007\000\008\000\008\000\008\000\008\000\008\000\
    \255\255\241\000\255\255\009\000\241\000\009\000\009\000\009\000\
    \009\000\255\255\255\255\255\255\009\000\009\000\255\255\009\000\
    \009\000\009\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \007\000\241\000\007\000\008\000\009\000\008\000\009\000\009\000\
    \009\000\009\000\009\000\255\255\004\001\255\255\011\000\004\001\
    \011\000\011\000\011\000\011\000\255\255\255\255\204\000\011\000\
    \011\000\255\255\011\000\011\000\011\000\255\255\255\255\255\255\
    \134\000\255\255\255\255\008\000\027\000\008\000\009\000\011\000\
    \009\000\011\000\011\000\011\000\011\000\011\000\255\255\255\255\
    \135\000\255\255\255\255\013\000\255\255\013\000\013\000\013\000\
    \013\000\255\255\255\255\241\000\013\000\013\000\255\255\013\000\
    \013\000\013\000\255\255\255\255\255\255\255\255\009\000\255\255\
    \009\000\011\000\011\000\011\000\013\000\255\255\013\000\013\000\
    \013\000\013\000\013\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\015\001\255\255\255\255\
    \255\255\255\255\217\000\199\000\255\255\217\000\255\255\016\001\
    \018\001\011\000\255\255\011\000\255\255\255\255\013\000\255\255\
    \013\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \015\000\255\255\015\000\015\000\015\000\015\000\004\001\217\000\
    \255\255\015\000\015\000\255\255\015\000\015\000\015\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\013\000\013\000\
    \013\000\015\000\255\255\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\255\255\015\000\217\000\015\000\015\000\255\255\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\255\255\015\000\019\000\015\000\019\000\019\000\
    \019\000\019\000\255\255\255\255\255\255\019\000\019\000\241\000\
    \019\000\019\000\019\000\234\000\234\000\234\000\234\000\234\000\
    \234\000\234\000\234\000\234\000\234\000\019\000\255\255\019\000\
    \019\000\019\000\019\000\019\000\255\255\255\255\255\255\021\000\
    \255\255\021\000\021\000\021\000\021\000\255\255\255\255\255\255\
    \021\000\021\000\004\001\021\000\021\000\021\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\019\000\
    \021\000\019\000\021\000\021\000\021\000\021\000\021\000\255\255\
    \255\255\255\255\025\000\255\255\025\000\025\000\025\000\025\000\
    \255\255\255\255\255\255\025\000\025\000\255\255\025\000\025\000\
    \025\000\255\255\255\255\255\255\255\255\255\255\255\255\019\000\
    \255\255\019\000\021\000\025\000\021\000\025\000\025\000\025\000\
    \025\000\025\000\115\000\115\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\115\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\115\000\115\000\115\000\115\000\115\000\
    \115\000\255\255\021\000\255\255\021\000\025\000\255\255\025\000\
    \217\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\026\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\115\000\115\000\115\000\115\000\115\000\
    \115\000\255\255\255\255\255\255\255\255\025\000\255\255\025\000\
    \026\000\026\000\255\255\026\000\026\000\026\000\026\000\255\255\
    \255\255\255\255\026\000\026\000\255\255\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\255\255\026\000\026\000\026\000\026\000\
    \026\000\236\000\236\000\236\000\236\000\236\000\236\000\236\000\
    \236\000\236\000\236\000\250\000\250\000\250\000\250\000\250\000\
    \250\000\250\000\250\000\250\000\250\000\255\255\255\255\255\255\
    \255\255\255\255\028\000\255\255\026\000\101\000\026\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\255\255\101\000\255\255\026\000\028\000\026\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\030\000\028\000\255\255\255\255\255\255\255\255\255\255\
    \030\000\255\255\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\255\255\255\255\255\255\
    \255\255\030\000\255\255\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\031\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\031\000\255\255\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\255\255\255\255\255\255\255\255\031\000\255\255\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\032\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\255\255\255\255\255\255\
    \255\255\032\000\255\255\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\255\255\
    \255\255\255\255\255\255\069\000\255\255\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\255\000\
    \255\000\255\000\255\000\255\000\255\000\255\000\255\000\255\000\
    \255\000\255\255\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\255\255\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\033\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\255\255\255\255\255\255\
    \255\255\033\000\255\255\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\099\000\255\255\255\255\
    \099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\255\255\255\255\255\255\255\255\099\000\255\255\
    \099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\099\000\099\000\099\000\099\000\099\000\099\000\
    \099\000\099\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\255\255\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\034\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\255\255\255\255\255\255\
    \255\255\034\000\255\255\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\126\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\126\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\122\000\122\000\
    \122\000\122\000\122\000\122\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\255\255\255\255\
    \255\255\255\255\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\255\255\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\035\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\255\255\255\255\255\255\
    \255\255\035\000\255\255\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\125\000\125\000\255\255\
    \255\255\125\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\125\000\255\255\125\000\255\255\
    \255\255\255\255\255\255\212\000\255\255\255\255\212\000\255\255\
    \255\255\255\255\255\255\255\255\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\255\255\255\255\
    \255\255\255\255\255\255\212\000\255\255\255\255\255\255\255\255\
    \212\000\255\255\255\255\212\000\255\255\255\255\255\255\255\255\
    \212\000\255\255\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\255\255\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\036\000\035\000\035\000\035\000\035\000\035\000\035\000\
    \035\000\035\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\212\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\255\255\255\255\255\255\
    \255\255\036\000\255\255\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\255\255\
    \255\255\255\255\255\255\154\000\125\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\255\255\
    \255\255\212\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\255\255\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\037\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\255\255\255\255\255\255\
    \255\255\037\000\255\255\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\255\255\
    \255\255\255\255\255\255\176\000\255\255\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\255\255\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\255\255\037\000\037\000\037\000\037\000\037\000\037\000\
    \037\000\037\000\038\000\255\255\038\000\038\000\038\000\038\000\
    \255\255\255\255\255\255\038\000\038\000\255\255\038\000\038\000\
    \038\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\038\000\255\255\038\000\038\000\038\000\
    \038\000\038\000\255\255\255\255\255\255\039\000\255\255\039\000\
    \039\000\039\000\039\000\255\255\255\255\255\255\039\000\039\000\
    \255\255\039\000\039\000\039\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\038\000\039\000\038\000\
    \039\000\039\000\039\000\039\000\039\000\255\255\255\255\255\255\
    \040\000\255\255\040\000\040\000\040\000\040\000\255\255\255\255\
    \255\255\040\000\040\000\255\255\040\000\040\000\040\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\038\000\255\255\038\000\
    \039\000\040\000\039\000\040\000\040\000\040\000\040\000\040\000\
    \255\255\255\255\255\255\255\255\041\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\255\255\
    \039\000\255\255\039\000\040\000\255\255\040\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \255\255\255\255\255\255\040\000\041\000\040\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\255\255\255\255\255\255\255\255\178\000\255\255\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\255\255\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\255\255\041\000\041\000\041\000\
    \041\000\041\000\041\000\041\000\041\000\045\000\255\255\255\255\
    \045\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\045\000\255\255\255\255\045\000\
    \255\255\045\000\255\255\255\255\255\255\045\000\045\000\255\255\
    \045\000\255\255\045\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\045\000\
    \045\000\045\000\045\000\045\000\255\255\255\255\255\255\046\000\
    \255\255\046\000\046\000\046\000\046\000\255\255\255\255\255\255\
    \046\000\046\000\255\255\046\000\046\000\046\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \046\000\045\000\046\000\046\000\046\000\046\000\046\000\255\255\
    \255\255\255\255\047\000\255\255\047\000\047\000\047\000\047\000\
    \255\255\255\255\255\255\047\000\047\000\255\255\047\000\047\000\
    \047\000\255\255\255\255\255\255\255\255\255\255\255\255\045\000\
    \255\255\045\000\046\000\047\000\046\000\047\000\047\000\047\000\
    \047\000\047\000\255\255\255\255\255\255\048\000\255\255\048\000\
    \048\000\048\000\048\000\255\255\255\255\255\255\048\000\048\000\
    \255\255\048\000\048\000\048\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\046\000\255\255\046\000\047\000\048\000\047\000\
    \048\000\048\000\048\000\048\000\048\000\255\255\255\255\255\255\
    \049\000\255\255\049\000\049\000\049\000\049\000\255\255\255\255\
    \255\255\049\000\049\000\255\255\049\000\049\000\049\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\047\000\255\255\047\000\
    \048\000\049\000\048\000\049\000\049\000\049\000\049\000\049\000\
    \255\255\255\255\255\255\050\000\255\255\050\000\050\000\050\000\
    \050\000\255\255\255\255\255\255\050\000\050\000\255\255\050\000\
    \050\000\050\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \048\000\255\255\048\000\049\000\050\000\049\000\050\000\050\000\
    \050\000\050\000\050\000\255\255\255\255\255\255\053\000\255\255\
    \053\000\053\000\053\000\053\000\255\255\255\255\255\255\053\000\
    \053\000\255\255\053\000\053\000\053\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\049\000\255\255\049\000\050\000\053\000\
    \050\000\053\000\053\000\053\000\053\000\053\000\255\255\255\255\
    \255\255\055\000\255\255\055\000\055\000\055\000\055\000\255\255\
    \255\255\255\255\055\000\055\000\255\255\055\000\055\000\055\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\050\000\255\255\
    \050\000\053\000\055\000\053\000\055\000\055\000\055\000\055\000\
    \055\000\255\255\255\255\255\255\056\000\255\255\056\000\056\000\
    \056\000\056\000\255\255\255\255\255\255\056\000\056\000\255\255\
    \056\000\056\000\056\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\053\000\255\255\053\000\055\000\056\000\055\000\056\000\
    \056\000\056\000\056\000\056\000\255\255\255\255\255\255\058\000\
    \255\255\058\000\058\000\058\000\058\000\255\255\255\255\255\255\
    \058\000\058\000\255\255\058\000\058\000\058\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\055\000\255\255\055\000\056\000\
    \058\000\056\000\058\000\058\000\058\000\058\000\058\000\255\255\
    \233\000\233\000\233\000\233\000\233\000\233\000\233\000\233\000\
    \233\000\233\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\233\000\233\000\233\000\233\000\233\000\233\000\056\000\
    \255\255\056\000\058\000\255\255\058\000\255\255\255\255\255\255\
    \255\255\255\255\064\000\255\255\064\000\064\000\064\000\064\000\
    \255\255\255\255\255\255\064\000\064\000\255\255\064\000\064\000\
    \064\000\233\000\233\000\233\000\233\000\233\000\233\000\255\255\
    \255\255\255\255\058\000\064\000\058\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\255\255\064\000\255\255\064\000\
    \064\000\255\255\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\255\255\064\000\065\000\064\000\
    \065\000\065\000\065\000\065\000\255\255\255\255\255\255\065\000\
    \065\000\255\255\065\000\065\000\065\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\065\000\
    \255\255\065\000\065\000\065\000\065\000\065\000\255\255\255\255\
    \255\255\066\000\255\255\066\000\066\000\066\000\066\000\255\255\
    \255\255\255\255\066\000\066\000\255\255\066\000\066\000\066\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\065\000\066\000\065\000\066\000\066\000\066\000\066\000\
    \066\000\255\255\238\000\238\000\238\000\238\000\238\000\238\000\
    \238\000\238\000\238\000\238\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\238\000\238\000\238\000\238\000\238\000\
    \238\000\065\000\255\255\065\000\066\000\255\255\066\000\255\255\
    \255\255\255\255\255\255\255\255\067\000\255\255\067\000\067\000\
    \067\000\067\000\255\255\255\255\255\255\067\000\067\000\255\255\
    \067\000\067\000\067\000\238\000\238\000\238\000\238\000\238\000\
    \238\000\255\255\255\255\255\255\066\000\067\000\066\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\255\255\067\000\
    \255\255\067\000\067\000\255\255\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\068\000\067\000\
    \255\255\067\000\255\255\255\255\255\255\068\000\255\255\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\255\255\255\255\255\255\255\255\068\000\255\255\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\070\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\255\255\255\255\
    \255\255\255\255\070\000\255\255\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \255\255\255\255\255\255\255\255\186\000\255\255\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \186\000\186\000\186\000\186\000\186\000\186\000\186\000\186\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\071\000\255\255\255\255\071\000\255\255\
    \071\000\071\000\255\255\255\255\071\000\071\000\255\255\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\255\255\255\255\071\000\071\000\
    \071\000\255\255\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\255\255\255\255\255\255\
    \071\000\071\000\255\255\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\218\000\071\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\225\000\218\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\225\000\225\000\255\255\225\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\255\255\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\255\255\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\072\000\255\255\255\255\072\000\072\000\072\000\
    \255\255\255\255\255\255\072\000\072\000\255\255\072\000\255\255\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\255\255\255\255\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\255\255\072\000\255\255\072\000\
    \072\000\255\255\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\073\000\072\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \255\255\255\255\255\255\255\255\073\000\255\255\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \249\000\249\000\249\000\249\000\249\000\249\000\249\000\249\000\
    \249\000\249\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\249\000\249\000\249\000\249\000\249\000\249\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\249\000\249\000\249\000\249\000\249\000\249\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\255\255\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\255\255\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\074\000\255\255\255\255\
    \074\000\255\255\074\000\074\000\255\255\255\255\074\000\074\000\
    \255\255\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\255\255\255\255\
    \074\000\074\000\074\000\255\255\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\255\255\
    \255\255\255\255\074\000\074\000\255\255\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\255\255\
    \074\000\001\001\001\001\001\001\001\001\001\001\001\001\001\001\
    \001\001\001\001\001\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\001\001\001\001\001\001\001\001\001\001\001\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\001\001\001\001\001\001\001\001\001\001\001\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\255\255\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\255\255\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\075\000\255\255\255\255\075\000\
    \075\000\075\000\255\255\255\255\255\255\075\000\075\000\255\255\
    \075\000\255\255\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\255\255\255\255\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\255\255\075\000\
    \255\255\075\000\075\000\255\255\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\076\000\075\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\255\255\255\255\255\255\255\255\076\000\255\255\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\077\000\255\255\255\255\255\255\077\000\255\255\
    \077\000\255\255\255\255\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\255\255\255\255\
    \255\255\255\255\077\000\255\255\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\078\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\255\255\255\255\255\255\255\255\078\000\255\255\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\079\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\255\255\255\255\
    \255\255\255\255\079\000\255\255\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\080\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\255\255\255\255\255\255\255\255\080\000\255\255\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\081\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\255\255\255\255\
    \255\255\255\255\081\000\255\255\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\082\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\255\255\255\255\255\255\255\255\082\000\255\255\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\083\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\083\000\255\255\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\255\255\255\255\
    \255\255\255\255\083\000\255\255\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\084\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\255\255\255\255\255\255\255\255\084\000\255\255\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\085\000\255\255\255\255\255\255\085\000\255\255\
    \085\000\255\255\255\255\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\255\255\255\255\
    \255\255\255\255\085\000\255\255\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\086\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\255\255\255\255\255\255\255\255\086\000\255\255\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\088\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\255\255\255\255\
    \255\255\255\255\088\000\255\255\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\089\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\255\255\255\255\255\255\255\255\089\000\255\255\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\090\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\255\255\255\255\
    \255\255\255\255\090\000\255\255\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\091\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\255\255\255\255\255\255\255\255\091\000\255\255\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\091\000\092\000\255\255\255\255\255\255\092\000\255\255\
    \092\000\255\255\255\255\092\000\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\092\000\255\255\255\255\
    \255\255\255\255\092\000\255\255\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\092\000\093\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\093\000\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\093\000\255\255\255\255\255\255\255\255\093\000\255\255\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\093\000\093\000\093\000\093\000\093\000\093\000\093\000\
    \093\000\093\000\094\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\094\000\094\000\094\000\094\000\094\000\
    \094\000\094\000\094\000\094\000\094\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\094\000\094\000\094\000\094\000\
    \094\000\094\000\094\000\094\000\094\000\094\000\094\000\094\000\
    \094\000\094\000\094\000\094\000\094\000\094\000\094\000\094\000\
    \094\000\094\000\094\000\094\000\094\000\094\000\255\255\255\255\
    \255\255\255\255\094\000\255\255\094\000\094\000\094\000\094\000\
    \094\000\094\000\094\000\094\000\094\000\094\000\094\000\094\000\
    \094\000\094\000\094\000\094\000\094\000\094\000\094\000\094\000\
    \094\000\094\000\094\000\094\000\094\000\094\000\095\000\255\255\
    \255\255\255\255\095\000\255\255\095\000\255\255\255\255\095\000\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \095\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \095\000\095\000\255\255\255\255\255\255\255\255\095\000\255\255\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \095\000\095\000\096\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\255\255\255\255\
    \255\255\255\255\096\000\255\255\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\098\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\098\000\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\098\000\255\255\255\255\255\255\255\255\098\000\255\255\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \098\000\098\000\103\000\255\255\255\255\103\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\103\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \103\000\255\255\255\255\255\255\255\255\255\255\255\255\103\000\
    \255\255\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\255\255\255\255\255\255\255\255\
    \103\000\255\255\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\104\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\104\000\255\255\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \255\255\255\255\255\255\255\255\104\000\255\255\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \108\000\255\255\108\000\255\255\255\255\255\255\255\255\108\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\
    \108\000\108\000\127\000\255\255\127\000\127\000\127\000\127\000\
    \255\255\255\255\255\255\127\000\127\000\255\255\127\000\127\000\
    \127\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\127\000\255\255\127\000\127\000\127\000\
    \127\000\127\000\255\255\255\255\108\000\255\255\255\255\255\255\
    \255\255\255\255\108\000\129\000\255\255\129\000\129\000\129\000\
    \129\000\255\255\255\255\255\255\129\000\129\000\108\000\129\000\
    \129\000\129\000\108\000\255\255\108\000\127\000\255\255\127\000\
    \108\000\255\255\255\255\255\255\129\000\255\255\129\000\129\000\
    \129\000\129\000\129\000\255\255\255\255\255\255\130\000\255\255\
    \130\000\130\000\130\000\130\000\255\255\255\255\255\255\130\000\
    \130\000\255\255\130\000\130\000\130\000\127\000\255\255\127\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\129\000\130\000\
    \129\000\130\000\130\000\130\000\130\000\130\000\255\255\255\255\
    \255\255\139\000\255\255\139\000\139\000\139\000\139\000\255\255\
    \255\255\255\255\139\000\139\000\255\255\139\000\139\000\139\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\129\000\255\255\
    \129\000\130\000\139\000\130\000\139\000\139\000\139\000\139\000\
    \139\000\255\255\255\255\255\255\140\000\255\255\140\000\140\000\
    \140\000\140\000\255\255\255\255\255\255\140\000\140\000\255\255\
    \140\000\140\000\140\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\130\000\255\255\130\000\139\000\140\000\139\000\140\000\
    \140\000\140\000\140\000\140\000\255\255\255\255\255\255\141\000\
    \255\255\141\000\141\000\141\000\141\000\255\255\255\255\255\255\
    \141\000\141\000\255\255\141\000\141\000\141\000\255\255\255\255\
    \108\000\255\255\255\255\255\255\139\000\255\255\139\000\140\000\
    \141\000\140\000\141\000\141\000\141\000\141\000\141\000\255\255\
    \255\255\255\255\148\000\255\255\148\000\148\000\148\000\148\000\
    \255\255\255\255\255\255\148\000\148\000\255\255\148\000\148\000\
    \148\000\255\255\255\255\255\255\255\255\255\255\255\255\140\000\
    \255\255\140\000\141\000\148\000\141\000\148\000\148\000\148\000\
    \148\000\148\000\255\255\255\255\255\255\149\000\255\255\149\000\
    \149\000\149\000\149\000\255\255\255\255\255\255\149\000\149\000\
    \255\255\149\000\149\000\149\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\141\000\255\255\141\000\148\000\149\000\148\000\
    \149\000\149\000\149\000\149\000\149\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \150\000\255\255\255\255\150\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\148\000\255\255\148\000\
    \149\000\255\255\149\000\255\255\255\255\255\255\255\255\150\000\
    \150\000\255\255\150\000\150\000\150\000\150\000\255\255\255\255\
    \255\255\150\000\150\000\255\255\150\000\150\000\150\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \149\000\150\000\149\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\255\255\150\000\255\255\150\000\150\000\255\255\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\151\000\150\000\255\255\150\000\255\255\255\255\
    \255\255\255\255\255\255\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\255\255\255\255\
    \255\255\255\255\151\000\255\255\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\152\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\255\255\255\255\255\255\255\255\152\000\255\255\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\152\000\152\000\152\000\152\000\152\000\152\000\
    \152\000\152\000\153\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\153\000\255\255\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\255\255\255\255\
    \255\255\255\255\153\000\255\255\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\155\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\155\000\255\255\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\255\255\255\255\255\255\255\255\155\000\255\255\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\156\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\255\255\255\255\
    \255\255\255\255\156\000\255\255\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\157\000\255\255\
    \157\000\157\000\157\000\157\000\255\255\255\255\255\255\157\000\
    \157\000\255\255\157\000\157\000\157\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\157\000\
    \255\255\157\000\157\000\157\000\157\000\157\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\158\000\255\255\255\255\
    \158\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\157\000\255\255\157\000\158\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\158\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\158\000\255\255\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\255\255\255\255\
    \255\255\157\000\158\000\157\000\255\255\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\255\255\
    \255\255\255\255\255\255\158\000\255\255\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\159\000\
    \255\255\255\255\159\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\159\000\162\000\
    \255\255\162\000\162\000\162\000\162\000\255\255\255\255\255\255\
    \162\000\162\000\255\255\162\000\162\000\162\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \162\000\255\255\162\000\162\000\162\000\162\000\162\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\255\255\162\000\255\255\162\000\159\000\255\255\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\255\255\162\000\163\000\162\000\163\000\163\000\163\000\
    \163\000\255\255\255\255\255\255\163\000\163\000\255\255\163\000\
    \163\000\163\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\163\000\255\255\163\000\163\000\
    \163\000\163\000\163\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\163\000\255\255\
    \163\000\255\255\255\255\255\255\255\255\255\255\171\000\255\255\
    \171\000\171\000\171\000\171\000\255\255\255\255\255\255\171\000\
    \171\000\255\255\171\000\171\000\171\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\163\000\171\000\
    \163\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \255\255\171\000\255\255\171\000\171\000\255\255\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \255\255\171\000\173\000\171\000\173\000\173\000\173\000\173\000\
    \255\255\255\255\255\255\173\000\173\000\255\255\173\000\173\000\
    \173\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\173\000\255\255\173\000\173\000\173\000\
    \173\000\173\000\255\255\255\255\255\255\174\000\255\255\174\000\
    \174\000\174\000\174\000\255\255\255\255\255\255\174\000\174\000\
    \255\255\174\000\174\000\174\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\173\000\174\000\173\000\
    \174\000\174\000\174\000\174\000\174\000\255\255\255\255\255\255\
    \175\000\255\255\175\000\175\000\175\000\175\000\255\255\255\255\
    \255\255\175\000\175\000\255\255\175\000\175\000\175\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\173\000\255\255\173\000\
    \174\000\175\000\174\000\175\000\175\000\175\000\175\000\175\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \174\000\255\255\174\000\175\000\255\255\175\000\255\255\255\255\
    \255\255\255\255\255\255\177\000\255\255\177\000\177\000\177\000\
    \177\000\255\255\255\255\255\255\177\000\177\000\255\255\177\000\
    \177\000\177\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\175\000\177\000\175\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\255\255\177\000\255\255\
    \177\000\177\000\255\255\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\177\000\177\000\177\000\255\255\177\000\180\000\
    \177\000\180\000\180\000\180\000\180\000\255\255\255\255\255\255\
    \180\000\180\000\255\255\180\000\180\000\180\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \180\000\255\255\180\000\180\000\180\000\180\000\180\000\255\255\
    \255\255\255\255\181\000\255\255\181\000\181\000\181\000\181\000\
    \255\255\255\255\255\255\181\000\181\000\255\255\181\000\181\000\
    \181\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\180\000\181\000\180\000\181\000\181\000\181\000\
    \181\000\181\000\255\255\255\255\255\255\182\000\255\255\182\000\
    \182\000\182\000\182\000\255\255\255\255\255\255\182\000\182\000\
    \255\255\182\000\182\000\182\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\180\000\255\255\180\000\181\000\182\000\181\000\
    \182\000\182\000\182\000\182\000\182\000\255\255\255\255\255\255\
    \183\000\255\255\183\000\183\000\183\000\183\000\255\255\255\255\
    \255\255\183\000\183\000\255\255\183\000\183\000\183\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\181\000\255\255\181\000\
    \182\000\183\000\182\000\183\000\183\000\183\000\183\000\183\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \182\000\255\255\182\000\183\000\183\000\183\000\255\255\255\255\
    \255\255\255\255\255\255\185\000\255\255\185\000\185\000\185\000\
    \185\000\255\255\255\255\255\255\185\000\185\000\255\255\185\000\
    \185\000\185\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\183\000\185\000\183\000\185\000\185\000\
    \185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\
    \185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\
    \185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\
    \185\000\185\000\185\000\185\000\185\000\255\255\185\000\255\255\
    \185\000\185\000\255\255\185\000\185\000\185\000\185\000\185\000\
    \185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\
    \185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\
    \185\000\185\000\185\000\185\000\185\000\255\255\185\000\187\000\
    \185\000\187\000\187\000\187\000\187\000\255\255\255\255\255\255\
    \187\000\187\000\255\255\187\000\187\000\187\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \187\000\255\255\187\000\187\000\187\000\187\000\187\000\255\255\
    \255\255\255\255\188\000\255\255\188\000\188\000\188\000\188\000\
    \255\255\255\255\255\255\188\000\188\000\255\255\188\000\188\000\
    \188\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\187\000\188\000\187\000\188\000\188\000\188\000\
    \188\000\188\000\255\255\255\255\255\255\189\000\255\255\189\000\
    \189\000\189\000\189\000\255\255\255\255\255\255\189\000\189\000\
    \255\255\189\000\189\000\189\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\187\000\255\255\187\000\188\000\189\000\188\000\
    \189\000\189\000\189\000\189\000\189\000\255\255\255\255\255\255\
    \190\000\255\255\190\000\190\000\190\000\190\000\255\255\255\255\
    \255\255\190\000\190\000\255\255\190\000\190\000\190\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\188\000\255\255\188\000\
    \189\000\190\000\189\000\190\000\190\000\190\000\190\000\190\000\
    \255\255\255\255\255\255\191\000\255\255\191\000\191\000\191\000\
    \191\000\255\255\255\255\255\255\191\000\191\000\255\255\191\000\
    \191\000\191\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \189\000\255\255\189\000\190\000\191\000\190\000\191\000\191\000\
    \191\000\191\000\191\000\255\255\255\255\255\255\194\000\255\255\
    \194\000\194\000\194\000\194\000\255\255\255\255\255\255\194\000\
    \194\000\255\255\194\000\194\000\194\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\190\000\255\255\190\000\191\000\194\000\
    \191\000\194\000\194\000\194\000\194\000\194\000\255\255\255\255\
    \255\255\197\000\255\255\197\000\197\000\197\000\197\000\255\255\
    \255\255\255\255\197\000\197\000\255\255\197\000\197\000\197\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\191\000\255\255\
    \191\000\194\000\197\000\194\000\197\000\197\000\197\000\197\000\
    \197\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\226\000\
    \255\255\226\000\255\255\255\255\255\255\255\255\226\000\255\255\
    \255\255\194\000\246\000\194\000\197\000\246\000\197\000\226\000\
    \226\000\226\000\226\000\226\000\226\000\226\000\226\000\226\000\
    \226\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \246\000\255\255\246\000\255\255\255\255\255\255\255\255\246\000\
    \255\255\255\255\255\255\255\255\197\000\255\255\197\000\255\255\
    \246\000\246\000\246\000\246\000\246\000\246\000\246\000\246\000\
    \246\000\246\000\255\255\226\000\255\255\255\255\255\255\255\255\
    \255\255\226\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\226\000\255\255\255\255\
    \255\255\226\000\255\255\226\000\255\255\255\255\255\255\226\000\
    \255\255\255\255\255\255\255\255\246\000\255\255\255\255\255\255\
    \255\255\255\255\246\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\246\000\255\255\
    \255\255\255\255\246\000\255\255\246\000\255\255\255\255\006\001\
    \246\000\006\001\006\001\006\001\006\001\006\001\006\001\006\001\
    \006\001\006\001\006\001\006\001\006\001\006\001\006\001\006\001\
    \006\001\006\001\006\001\006\001\006\001\006\001\006\001\006\001\
    \006\001\006\001\006\001\255\255\012\001\006\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \012\001\012\001\012\001\012\001\012\001\012\001\012\001\012\001\
    \255\255\255\255\012\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \246\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\010\000\000\000\000\000\000\000\022\000\034\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\044\000\054\000\000\000\
    \064\000\070\000\085\000\110\000\165\000\140\000\188\000\198\000\
    \220\000\000\000\000\000\000\000\236\000\246\000\078\000\254\000\
    \012\001\022\001\038\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\048\001\074\001\000\000\
    \000\000\000\000\000\000\029\000\000\000\000\000\000\000\002\000\
    \000\000\044\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\100\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\184\001\042\002\
    \056\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\001\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\002\000\004\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\018\000\018\000\000\000\000\000\
    \018\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\018\000\000\000\000\000\000\000\000\000\000\000\018\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\064\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\036\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\061\000\061\000\000\000\061\000\000\000\
    \000\000\000\000\000\000\001\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\001\000\039\000\000\000\000\000\000\000\
    \000\000\000\000\001\000\000\000\000\000\000\000\000\000\009\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\015\000\000\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \015\000\000\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \021\000\021\000\000\000\000\000\000\000\004\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\021\000\021\000\
    \000\000\004\000\000\000\000\000\000\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\015\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\015\000\000\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \000\000\000\000\000\000\000\000\000\000\021\000\000\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\000\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\004\000\000\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \000\000\000\000\000\000\000\000\015\000\000\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\000\000\000\000\
    \000\000\000\000\000\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\021\000\021\000\
    \021\000\021\000\021\000\021\000\021\000\021\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \000\000\000\000\000\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \000\000\000\000\024\000\001\000\000\000\021\000\015\000\015\000\
    \015\000\015\000\015\000\015\000\015\000\015\000\015\000\015\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\001\000\015\000\000\000\009\000\000\000\000\000\
    \009\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\009\000\015\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\000\000\
    \000\000\000\000\000\000\053\000\000\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\053\000\000\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\000\000\000\000\000\000\000\000\053\000\
    \000\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\009\000\000\000\000\000\009\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\009\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\000\000\000\000\000\000\
    \000\000\053\000\000\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\134\000\246\000\253\000\134\000\254\000\255\255\
    \255\255\255\255\255\255\026\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\134\000\000\000\135\000\255\255\255\255\255\255\
    \255\255\255\255\026\000\255\255\255\255\255\255\255\255\015\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\030\000\255\255\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \031\000\255\255\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \080\000\080\000\255\255\255\255\255\255\030\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\094\000\094\000\
    \255\255\031\000\255\255\255\255\255\255\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\078\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\083\000\255\255\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \255\255\255\255\255\255\255\255\255\255\094\000\255\255\083\000\
    \083\000\083\000\083\000\083\000\083\000\255\255\082\000\082\000\
    \082\000\082\000\082\000\082\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\083\000\255\255\083\000\
    \083\000\083\000\083\000\083\000\083\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\084\000\084\000\
    \084\000\084\000\084\000\084\000\086\000\086\000\086\000\086\000\
    \086\000\086\000\086\000\086\000\086\000\086\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\087\000\
    \134\000\255\255\255\255\255\255\084\000\255\255\084\000\084\000\
    \084\000\084\000\084\000\084\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\255\255\255\255\
    \255\255\255\255\255\255\086\000\092\000\092\000\092\000\092\000\
    \092\000\092\000\092\000\092\000\092\000\092\000\093\000\093\000\
    \093\000\093\000\093\000\093\000\093\000\093\000\095\000\095\000\
    \095\000\095\000\095\000\095\000\095\000\095\000\095\000\095\000\
    \255\255\255\255\255\255\088\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\097\000\097\000\
    \097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\
    \255\255\255\255\125\000\126\000\255\255\093\000\098\000\098\000\
    \098\000\098\000\098\000\098\000\098\000\098\000\098\000\098\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\126\000\096\000\255\255\150\000\255\255\255\255\
    \150\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\150\000\098\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\255\255\
    \255\255\255\255\255\255\150\000\255\255\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\158\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\158\000\255\255\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\255\255\255\255\255\255\255\255\158\000\
    \255\255\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\159\000\255\255\255\255\159\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\159\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\255\255\255\255\255\255\
    \255\255\159\000\255\255\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\159\000\159\000\159\000\
    \159\000\159\000\159\000\159\000\159\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_code =
   "\255\004\255\255\006\255\005\255\255\010\255\255\007\255\255\005\
    \255\255\000\005\255\006\255\255\009\255\008\255\255\003\255\000\
    \004\001\007\255\009\255\255\008\255\009\255\255\000\004\001\007\
    \003\008\002\009\255\011\255\255\000\010\001\011\255\001\255\255\
    \000\001\255";
}

let rec token state lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 12 (-1); __ocaml_lex_token_rec state lexbuf 0
and __ocaml_lex_token_rec state lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 353 "src/reason-parser/reason_declarative_lexer.mll"
                 (
      raise_error
        (Location.curr lexbuf)
        (Illegal_character (Lexing.lexeme_char lexbuf 0));
      update_loc lexbuf None 1 false 0;
      token state lexbuf
    )
# 2906 "src/reason-parser/reason_declarative_lexer.ml"

  | 1 ->
# 361 "src/reason-parser/reason_declarative_lexer.mll"
    ( update_loc lexbuf None 1 false 0;
      token state lexbuf
    )
# 2913 "src/reason-parser/reason_declarative_lexer.ml"

  | 2 ->
# 365 "src/reason-parser/reason_declarative_lexer.mll"
    ( token state lexbuf )
# 2918 "src/reason-parser/reason_declarative_lexer.ml"

  | 3 ->
# 367 "src/reason-parser/reason_declarative_lexer.mll"
    ( UNDERSCORE )
# 2923 "src/reason-parser/reason_declarative_lexer.ml"

  | 4 ->
# 369 "src/reason-parser/reason_declarative_lexer.mll"
    ( TILDE )
# 2928 "src/reason-parser/reason_declarative_lexer.ml"

  | 5 ->
# 371 "src/reason-parser/reason_declarative_lexer.mll"
    ( QUESTION )
# 2933 "src/reason-parser/reason_declarative_lexer.ml"

  | 6 ->
# 373 "src/reason-parser/reason_declarative_lexer.mll"
    ( set_lexeme_length lexbuf 1; EQUAL )
# 2938 "src/reason-parser/reason_declarative_lexer.ml"

  | 7 ->
# 375 "src/reason-parser/reason_declarative_lexer.mll"
    ( let s = Lexing.lexeme lexbuf in
      try Hashtbl.find keyword_table s
      with Not_found -> LIDENT s
    )
# 2946 "src/reason-parser/reason_declarative_lexer.ml"

  | 8 ->
# 380 "src/reason-parser/reason_declarative_lexer.mll"
    ( Ocaml_util.warn_latin1 lexbuf; LIDENT (Lexing.lexeme lexbuf) )
# 2951 "src/reason-parser/reason_declarative_lexer.ml"

  | 9 ->
# 382 "src/reason-parser/reason_declarative_lexer.mll"
    ( UIDENT(Lexing.lexeme lexbuf) )
# 2956 "src/reason-parser/reason_declarative_lexer.ml"

  | 10 ->
# 384 "src/reason-parser/reason_declarative_lexer.mll"
    ( Ocaml_util.warn_latin1 lexbuf; UIDENT(Lexing.lexeme lexbuf) )
# 2961 "src/reason-parser/reason_declarative_lexer.ml"

  | 11 ->
# 386 "src/reason-parser/reason_declarative_lexer.mll"
    ( INT (Lexing.lexeme lexbuf, None) )
# 2966 "src/reason-parser/reason_declarative_lexer.ml"

  | 12 ->
let
# 387 "src/reason-parser/reason_declarative_lexer.mll"
                    lit
# 2972 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_curr_pos + -1)
and
# 387 "src/reason-parser/reason_declarative_lexer.mll"
                                              modif
# 2977 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_curr_pos + -1) in
# 388 "src/reason-parser/reason_declarative_lexer.mll"
    ( INT (lit, Some modif) )
# 2981 "src/reason-parser/reason_declarative_lexer.ml"

  | 13 ->
# 390 "src/reason-parser/reason_declarative_lexer.mll"
    ( FLOAT (Lexing.lexeme lexbuf, None) )
# 2986 "src/reason-parser/reason_declarative_lexer.ml"

  | 14 ->
let
# 391 "src/reason-parser/reason_declarative_lexer.mll"
                                            lit
# 2992 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_curr_pos + -1)
and
# 391 "src/reason-parser/reason_declarative_lexer.mll"
                                                                      modif
# 2997 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_curr_pos + -1) in
# 392 "src/reason-parser/reason_declarative_lexer.mll"
    ( FLOAT (lit, Some modif) )
# 3001 "src/reason-parser/reason_declarative_lexer.ml"

  | 15 ->
let
# 393 "src/reason-parser/reason_declarative_lexer.mll"
                                            lit
# 3007 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(0) in
# 394 "src/reason-parser/reason_declarative_lexer.mll"
    ( raise_error
        (Location.curr lexbuf)
        (Invalid_literal (Lexing.lexeme lexbuf));
      FLOAT (lit, None)
    )
# 3015 "src/reason-parser/reason_declarative_lexer.ml"

  | 16 ->
let
# 399 "src/reason-parser/reason_declarative_lexer.mll"
                    lit
# 3021 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(0) in
# 400 "src/reason-parser/reason_declarative_lexer.mll"
    ( raise_error
        (Location.curr lexbuf)
        (Invalid_literal (Lexing.lexeme lexbuf));
      INT (lit, None)
    )
# 3029 "src/reason-parser/reason_declarative_lexer.ml"

  | 17 ->
# 406 "src/reason-parser/reason_declarative_lexer.mll"
    ( let string_start = lexbuf.lex_start_p in
      let start_loc = Location.curr lexbuf in
      let raw_buffer, txt_buffer = get_scratch_buffers state in
      if not (string raw_buffer (Some txt_buffer) lexbuf) then
        raise_error start_loc Unterminated_string;
      lexbuf.lex_start_p <- string_start;
      let txt = flush_buffer txt_buffer in
      let raw = flush_buffer raw_buffer in
      STRING (txt, Some raw, None)
    )
# 3043 "src/reason-parser/reason_declarative_lexer.ml"

  | 18 ->
let
# 416 "src/reason-parser/reason_declarative_lexer.mll"
                       delim
# 3049 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 417 "src/reason-parser/reason_declarative_lexer.mll"
    ( let string_start = lexbuf.lex_start_p in
      let start_loc = Location.curr lexbuf in
      let raw_buffer, _ = get_scratch_buffers state in
      if not (quoted_string raw_buffer delim lexbuf) then
        raise_error start_loc Unterminated_string;
      lexbuf.lex_start_p <- string_start;
      let txt = flush_buffer raw_buffer in
      STRING (txt, None, Some delim)
    )
# 3061 "src/reason-parser/reason_declarative_lexer.ml"

  | 19 ->
# 427 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* newline can span multiple characters
         (if the newline starts with \13)
         Only the first one is returned, maybe we should warn? *)
      update_loc lexbuf None 1 false 1;
      CHAR (Lexing.lexeme_char lexbuf 1)
    )
# 3071 "src/reason-parser/reason_declarative_lexer.ml"

  | 20 ->
let
# 433 "src/reason-parser/reason_declarative_lexer.mll"
                                        c
# 3077 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1) in
# 434 "src/reason-parser/reason_declarative_lexer.mll"
    ( CHAR c )
# 3081 "src/reason-parser/reason_declarative_lexer.ml"

  | 21 ->
let
# 435 "src/reason-parser/reason_declarative_lexer.mll"
                                                  c
# 3087 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 436 "src/reason-parser/reason_declarative_lexer.mll"
    ( CHAR (char_for_backslash c) )
# 3091 "src/reason-parser/reason_declarative_lexer.ml"

  | 22 ->
# 438 "src/reason-parser/reason_declarative_lexer.mll"
    ( CHAR (char_for_decimal_code lexbuf 2) )
# 3096 "src/reason-parser/reason_declarative_lexer.ml"

  | 23 ->
# 440 "src/reason-parser/reason_declarative_lexer.mll"
    ( CHAR (char_for_hexadecimal_code lexbuf 3) )
# 3101 "src/reason-parser/reason_declarative_lexer.ml"

  | 24 ->
let
# 441 "src/reason-parser/reason_declarative_lexer.mll"
                     esc
# 3107 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_start_pos + 3) in
# 442 "src/reason-parser/reason_declarative_lexer.mll"
    ( raise_error (Location.curr lexbuf) (Illegal_escape esc);
      token state lexbuf
    )
# 3113 "src/reason-parser/reason_declarative_lexer.ml"

  | 25 ->
# 446 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* Allow parsing of foo#=<bar /> *)
      set_lexeme_length lexbuf 2;
      SHARPEQUAL
    )
# 3121 "src/reason-parser/reason_declarative_lexer.ml"

  | 26 ->
# 451 "src/reason-parser/reason_declarative_lexer.mll"
    ( SHARPEQUAL )
# 3126 "src/reason-parser/reason_declarative_lexer.ml"

  | 27 ->
# 453 "src/reason-parser/reason_declarative_lexer.mll"
    ( SHARPOP (lexeme_operator lexbuf) )
# 3131 "src/reason-parser/reason_declarative_lexer.ml"

  | 28 ->
let
# 454 "src/reason-parser/reason_declarative_lexer.mll"
                                   num
# 3137 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 455 "src/reason-parser/reason_declarative_lexer.mll"
                                           name
# 3142 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_mem.(2) in
# 457 "src/reason-parser/reason_declarative_lexer.mll"
    ( update_loc lexbuf name (int_of_string num) true 0;
      token state lexbuf
    )
# 3148 "src/reason-parser/reason_declarative_lexer.ml"

  | 29 ->
# 460 "src/reason-parser/reason_declarative_lexer.mll"
         ( AMPERSAND )
# 3153 "src/reason-parser/reason_declarative_lexer.ml"

  | 30 ->
# 461 "src/reason-parser/reason_declarative_lexer.mll"
         ( AMPERAMPER )
# 3158 "src/reason-parser/reason_declarative_lexer.ml"

  | 31 ->
# 462 "src/reason-parser/reason_declarative_lexer.mll"
         ( BACKQUOTE )
# 3163 "src/reason-parser/reason_declarative_lexer.ml"

  | 32 ->
# 463 "src/reason-parser/reason_declarative_lexer.mll"
         ( QUOTE )
# 3168 "src/reason-parser/reason_declarative_lexer.ml"

  | 33 ->
# 464 "src/reason-parser/reason_declarative_lexer.mll"
         ( LPAREN )
# 3173 "src/reason-parser/reason_declarative_lexer.ml"

  | 34 ->
# 465 "src/reason-parser/reason_declarative_lexer.mll"
         ( RPAREN )
# 3178 "src/reason-parser/reason_declarative_lexer.ml"

  | 35 ->
# 466 "src/reason-parser/reason_declarative_lexer.mll"
         ( STAR )
# 3183 "src/reason-parser/reason_declarative_lexer.ml"

  | 36 ->
# 467 "src/reason-parser/reason_declarative_lexer.mll"
         ( COMMA )
# 3188 "src/reason-parser/reason_declarative_lexer.ml"

  | 37 ->
# 468 "src/reason-parser/reason_declarative_lexer.mll"
         ( MINUSGREATER )
# 3193 "src/reason-parser/reason_declarative_lexer.ml"

  | 38 ->
# 469 "src/reason-parser/reason_declarative_lexer.mll"
         ( EQUALGREATER )
# 3198 "src/reason-parser/reason_declarative_lexer.ml"

  | 39 ->
# 471 "src/reason-parser/reason_declarative_lexer.mll"
                                                     (
    set_lexeme_length lexbuf 2;
    EQUALGREATER
  )
# 3206 "src/reason-parser/reason_declarative_lexer.ml"

  | 40 ->
# 475 "src/reason-parser/reason_declarative_lexer.mll"
         ( SHARP )
# 3211 "src/reason-parser/reason_declarative_lexer.ml"

  | 41 ->
# 476 "src/reason-parser/reason_declarative_lexer.mll"
         ( DOT )
# 3216 "src/reason-parser/reason_declarative_lexer.ml"

  | 42 ->
# 477 "src/reason-parser/reason_declarative_lexer.mll"
         ( DOTDOT )
# 3221 "src/reason-parser/reason_declarative_lexer.ml"

  | 43 ->
# 478 "src/reason-parser/reason_declarative_lexer.mll"
         ( DOTDOTDOT )
# 3226 "src/reason-parser/reason_declarative_lexer.ml"

  | 44 ->
# 479 "src/reason-parser/reason_declarative_lexer.mll"
         ( COLON )
# 3231 "src/reason-parser/reason_declarative_lexer.ml"

  | 45 ->
# 480 "src/reason-parser/reason_declarative_lexer.mll"
         ( COLONCOLON )
# 3236 "src/reason-parser/reason_declarative_lexer.ml"

  | 46 ->
# 481 "src/reason-parser/reason_declarative_lexer.mll"
         ( COLONEQUAL )
# 3241 "src/reason-parser/reason_declarative_lexer.ml"

  | 47 ->
# 482 "src/reason-parser/reason_declarative_lexer.mll"
         ( COLONGREATER )
# 3246 "src/reason-parser/reason_declarative_lexer.ml"

  | 48 ->
# 483 "src/reason-parser/reason_declarative_lexer.mll"
         ( SEMI )
# 3251 "src/reason-parser/reason_declarative_lexer.ml"

  | 49 ->
# 484 "src/reason-parser/reason_declarative_lexer.mll"
         ( SEMISEMI )
# 3256 "src/reason-parser/reason_declarative_lexer.ml"

  | 50 ->
# 485 "src/reason-parser/reason_declarative_lexer.mll"
         ( LESS )
# 3261 "src/reason-parser/reason_declarative_lexer.ml"

  | 51 ->
# 486 "src/reason-parser/reason_declarative_lexer.mll"
         ( EQUAL )
# 3266 "src/reason-parser/reason_declarative_lexer.ml"

  | 52 ->
# 487 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACKET )
# 3271 "src/reason-parser/reason_declarative_lexer.ml"

  | 53 ->
# 488 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACKETBAR )
# 3276 "src/reason-parser/reason_declarative_lexer.ml"

  | 54 ->
# 489 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACKETLESS )
# 3281 "src/reason-parser/reason_declarative_lexer.ml"

  | 55 ->
# 490 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACKETGREATER )
# 3286 "src/reason-parser/reason_declarative_lexer.ml"

  | 56 ->
let
# 492 "src/reason-parser/reason_declarative_lexer.mll"
                                                                   tag
# 3292 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) lexbuf.Lexing.lex_curr_pos in
# 496 "src/reason-parser/reason_declarative_lexer.mll"
    ( LESSIDENT tag )
# 3296 "src/reason-parser/reason_declarative_lexer.ml"

  | 57 ->
let
# 497 "src/reason-parser/reason_declarative_lexer.mll"
                                   tag
# 3302 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) lexbuf.Lexing.lex_curr_pos in
# 498 "src/reason-parser/reason_declarative_lexer.mll"
    ( LESSUIDENT tag )
# 3306 "src/reason-parser/reason_declarative_lexer.ml"

  | 58 ->
# 499 "src/reason-parser/reason_declarative_lexer.mll"
           ( GREATERDOTDOTDOT )
# 3311 "src/reason-parser/reason_declarative_lexer.ml"

  | 59 ->
# 509 "src/reason-parser/reason_declarative_lexer.mll"
    ( set_lexeme_length lexbuf 2;
      LBRACELESS
    )
# 3318 "src/reason-parser/reason_declarative_lexer.ml"

  | 60 ->
# 513 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* allows parsing of `{<Text` in <Description term={<Text text="Age" />}>
         as correct jsx
       *)
      set_lexeme_length lexbuf 1;
      LBRACE
    )
# 3328 "src/reason-parser/reason_declarative_lexer.ml"

  | 61 ->
# 519 "src/reason-parser/reason_declarative_lexer.mll"
          (
    set_lexeme_length lexbuf 1;
    LBRACE
  )
# 3336 "src/reason-parser/reason_declarative_lexer.ml"

  | 62 ->
# 523 "src/reason-parser/reason_declarative_lexer.mll"
           (
    set_lexeme_length lexbuf 2;
    LBRACELESS
  )
# 3344 "src/reason-parser/reason_declarative_lexer.ml"

  | 63 ->
let
# 527 "src/reason-parser/reason_declarative_lexer.mll"
                                                               tag
# 3350 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1) in
# 528 "src/reason-parser/reason_declarative_lexer.mll"
    ( LESSSLASHIDENTGREATER tag )
# 3354 "src/reason-parser/reason_declarative_lexer.ml"

  | 64 ->
# 529 "src/reason-parser/reason_declarative_lexer.mll"
         ( RBRACKET )
# 3359 "src/reason-parser/reason_declarative_lexer.ml"

  | 65 ->
# 530 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACE )
# 3364 "src/reason-parser/reason_declarative_lexer.ml"

  | 66 ->
# 531 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACELESS )
# 3369 "src/reason-parser/reason_declarative_lexer.ml"

  | 67 ->
# 532 "src/reason-parser/reason_declarative_lexer.mll"
         ( BAR )
# 3374 "src/reason-parser/reason_declarative_lexer.ml"

  | 68 ->
# 533 "src/reason-parser/reason_declarative_lexer.mll"
         ( BARBAR )
# 3379 "src/reason-parser/reason_declarative_lexer.ml"

  | 69 ->
# 534 "src/reason-parser/reason_declarative_lexer.mll"
         ( BARRBRACKET )
# 3384 "src/reason-parser/reason_declarative_lexer.ml"

  | 70 ->
# 535 "src/reason-parser/reason_declarative_lexer.mll"
         ( GREATER )
# 3389 "src/reason-parser/reason_declarative_lexer.ml"

  | 71 ->
# 541 "src/reason-parser/reason_declarative_lexer.mll"
         ( RBRACE )
# 3394 "src/reason-parser/reason_declarative_lexer.ml"

  | 72 ->
# 542 "src/reason-parser/reason_declarative_lexer.mll"
         ( GREATERRBRACE )
# 3399 "src/reason-parser/reason_declarative_lexer.ml"

  | 73 ->
# 544 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* allow `let x=<div />;` *)
      set_lexeme_length lexbuf 1;
      EQUAL
    )
# 3407 "src/reason-parser/reason_declarative_lexer.ml"

  | 74 ->
# 549 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* jsx in arrays: [|<div />|]*)
      set_lexeme_length lexbuf 2;
      SLASHGREATER
    )
# 3415 "src/reason-parser/reason_declarative_lexer.ml"

  | 75 ->
# 554 "src/reason-parser/reason_declarative_lexer.mll"
    ( set_lexeme_length lexbuf 2;
      LBRACKETBAR
    )
# 3422 "src/reason-parser/reason_declarative_lexer.ml"

  | 76 ->
# 559 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* allow parsing of <div asd=1></div> *)
      set_lexeme_length lexbuf 2;
      SLASHGREATER
    )
# 3430 "src/reason-parser/reason_declarative_lexer.ml"

  | 77 ->
# 564 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* allow parsing of <div asd=1></div> *)
      set_lexeme_length lexbuf 1;
      GREATER
    )
# 3438 "src/reason-parser/reason_declarative_lexer.ml"

  | 78 ->
# 569 "src/reason-parser/reason_declarative_lexer.mll"
    ( (* allow parsing of <div><span> *)
      set_lexeme_length lexbuf 1;
      GREATER
    )
# 3446 "src/reason-parser/reason_declarative_lexer.ml"

  | 79 ->
# 573 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACKETAT )
# 3451 "src/reason-parser/reason_declarative_lexer.ml"

  | 80 ->
# 574 "src/reason-parser/reason_declarative_lexer.mll"
         ( LBRACKETPERCENT )
# 3456 "src/reason-parser/reason_declarative_lexer.ml"

  | 81 ->
# 575 "src/reason-parser/reason_declarative_lexer.mll"
          ( LBRACKETPERCENTPERCENT )
# 3461 "src/reason-parser/reason_declarative_lexer.ml"

  | 82 ->
# 576 "src/reason-parser/reason_declarative_lexer.mll"
         ( BANG )
# 3466 "src/reason-parser/reason_declarative_lexer.ml"

  | 83 ->
# 577 "src/reason-parser/reason_declarative_lexer.mll"
         ( INFIXOP0 "!=" )
# 3471 "src/reason-parser/reason_declarative_lexer.ml"

  | 84 ->
# 578 "src/reason-parser/reason_declarative_lexer.mll"
          ( INFIXOP0 "!==" )
# 3476 "src/reason-parser/reason_declarative_lexer.ml"

  | 85 ->
# 579 "src/reason-parser/reason_declarative_lexer.mll"
           ( INFIXOP0 "!=" )
# 3481 "src/reason-parser/reason_declarative_lexer.ml"

  | 86 ->
# 580 "src/reason-parser/reason_declarative_lexer.mll"
            ( INFIXOP0 "!==" )
# 3486 "src/reason-parser/reason_declarative_lexer.ml"

  | 87 ->
# 581 "src/reason-parser/reason_declarative_lexer.mll"
         ( PLUS )
# 3491 "src/reason-parser/reason_declarative_lexer.ml"

  | 88 ->
# 582 "src/reason-parser/reason_declarative_lexer.mll"
         ( PLUSDOT )
# 3496 "src/reason-parser/reason_declarative_lexer.ml"

  | 89 ->
# 583 "src/reason-parser/reason_declarative_lexer.mll"
         ( PLUSEQ )
# 3501 "src/reason-parser/reason_declarative_lexer.ml"

  | 90 ->
# 584 "src/reason-parser/reason_declarative_lexer.mll"
         ( MINUS )
# 3506 "src/reason-parser/reason_declarative_lexer.ml"

  | 91 ->
# 585 "src/reason-parser/reason_declarative_lexer.mll"
         ( MINUSDOT )
# 3511 "src/reason-parser/reason_declarative_lexer.ml"

  | 92 ->
# 586 "src/reason-parser/reason_declarative_lexer.mll"
         ( LESSGREATER )
# 3516 "src/reason-parser/reason_declarative_lexer.ml"

  | 93 ->
# 587 "src/reason-parser/reason_declarative_lexer.mll"
          ( LESSSLASHGREATER )
# 3521 "src/reason-parser/reason_declarative_lexer.ml"

  | 94 ->
# 588 "src/reason-parser/reason_declarative_lexer.mll"
           ( LESSDOTDOTGREATER )
# 3526 "src/reason-parser/reason_declarative_lexer.ml"

  | 95 ->
# 590 "src/reason-parser/reason_declarative_lexer.mll"
    ( PREFIXOP (lexeme_operator lexbuf) )
# 3531 "src/reason-parser/reason_declarative_lexer.ml"

  | 96 ->
# 592 "src/reason-parser/reason_declarative_lexer.mll"
    (
      (* See decompose_token in Reason_single_parser.ml for how let `x=-1` is lexed
       * and broken up into multiple tokens when necessary. *)
      INFIXOP0 (lexeme_operator lexbuf)
    )
# 3540 "src/reason-parser/reason_declarative_lexer.ml"

  | 97 ->
# 598 "src/reason-parser/reason_declarative_lexer.mll"
    ( INFIXOP1 (lexeme_operator lexbuf) )
# 3545 "src/reason-parser/reason_declarative_lexer.ml"

  | 98 ->
# 600 "src/reason-parser/reason_declarative_lexer.mll"
    ( match lexeme_without_comment lexbuf with
      | "^." | "^|" ->
        (* ^| is not an infix op in [|a^|] *)
        set_lexeme_length lexbuf
          (if Lexing.lexeme_char lexbuf 0 = '\\' then 2 else 1);
        POSTFIXOP "^"
      | "^" -> POSTFIXOP "^"
      | op -> INFIXOP1 (unescape_operator op)
    )
# 3558 "src/reason-parser/reason_declarative_lexer.ml"

  | 99 ->
# 610 "src/reason-parser/reason_declarative_lexer.mll"
    ( INFIXOP1 (lexeme_operator lexbuf) )
# 3563 "src/reason-parser/reason_declarative_lexer.ml"

  | 100 ->
# 612 "src/reason-parser/reason_declarative_lexer.mll"
    ( INFIXOP2 (lexeme_operator lexbuf) )
# 3568 "src/reason-parser/reason_declarative_lexer.ml"

  | 101 ->
# 614 "src/reason-parser/reason_declarative_lexer.mll"
         ( SLASHGREATER )
# 3573 "src/reason-parser/reason_declarative_lexer.ml"

  | 102 ->
# 625 "src/reason-parser/reason_declarative_lexer.mll"
    ( INFIXOP4 (lexeme_operator lexbuf) )
# 3578 "src/reason-parser/reason_declarative_lexer.ml"

  | 103 ->
# 626 "src/reason-parser/reason_declarative_lexer.mll"
        ( PERCENT )
# 3583 "src/reason-parser/reason_declarative_lexer.ml"

  | 104 ->
# 628 "src/reason-parser/reason_declarative_lexer.mll"
    ( match lexeme_operator lexbuf with
      | "" ->
          (* If the operator is empty, it means the lexeme is beginning
           * by a comment sequence: we let the comment lexer handle
           * the case. *)
          enter_comment state lexbuf
      | op -> INFIXOP3 op )
# 3594 "src/reason-parser/reason_declarative_lexer.ml"

  | 105 ->
# 636 "src/reason-parser/reason_declarative_lexer.mll"
    ( INFIXOP3 (lexeme_operator lexbuf) )
# 3599 "src/reason-parser/reason_declarative_lexer.ml"

  | 106 ->
# 638 "src/reason-parser/reason_declarative_lexer.mll"
    ( LETOP (Reason_syntax_util.expand_letop_identifier (lexeme_operator lexbuf)) )
# 3604 "src/reason-parser/reason_declarative_lexer.ml"

  | 107 ->
# 640 "src/reason-parser/reason_declarative_lexer.mll"
    ( ANDOP (Reason_syntax_util.expand_letop_identifier (lexeme_operator lexbuf)) )
# 3609 "src/reason-parser/reason_declarative_lexer.ml"

  | 108 ->
# 641 "src/reason-parser/reason_declarative_lexer.mll"
        ( EOF )
# 3614 "src/reason-parser/reason_declarative_lexer.ml"

  | 109 ->
# 643 "src/reason-parser/reason_declarative_lexer.mll"
    ( raise_error
        (Location.curr lexbuf)
        (Illegal_character (Lexing.lexeme_char lexbuf 0));
      token state lexbuf
    )
# 3623 "src/reason-parser/reason_declarative_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_token_rec state lexbuf __ocaml_lex_state

and enter_comment state lexbuf =
   __ocaml_lex_enter_comment_rec state lexbuf 199
and __ocaml_lex_enter_comment_rec state lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 650 "src/reason-parser/reason_declarative_lexer.mll"
                                line
# 3636 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 2) lexbuf.Lexing.lex_curr_pos in
# 651 "src/reason-parser/reason_declarative_lexer.mll"
    ( update_loc lexbuf None 1 false 0;
      let physical_loc = Location.curr lexbuf in
      let location = { physical_loc with
        loc_end = { physical_loc.loc_end with
          (* Don't track trailing `\n` in the location
           * 1| // comment
           * 2| let x = 1;
           * By omitting the `\n` at the end of line 1, the location of the
           * comment spans line 1. Otherwise the comment on line 1 would end
           * on the second line. The printer looks at the closing pos_lnum
           * location to interleave whitespace correct. It needs to align
           * with what we visually see (i.e. it ends on line 1) *)
          pos_lnum = physical_loc.loc_end.pos_lnum - 1;
          pos_cnum = physical_loc.loc_end.pos_cnum + 1;
      }} in
      COMMENT (line, location)
    )
# 3656 "src/reason-parser/reason_declarative_lexer.ml"

  | 1 ->
let
# 668 "src/reason-parser/reason_declarative_lexer.mll"
                            line
# 3662 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 2) lexbuf.Lexing.lex_curr_pos in
# 669 "src/reason-parser/reason_declarative_lexer.mll"
    ( update_loc lexbuf None 1 false 0;
      let physical_loc = Location.curr lexbuf in
      let location = { physical_loc with
        loc_end = { physical_loc.loc_end with
          pos_lnum = physical_loc.loc_end.pos_lnum - 1;
          pos_cnum = physical_loc.loc_end.pos_cnum + 1;
      }} in
      COMMENT (line, location)
    )
# 3674 "src/reason-parser/reason_declarative_lexer.ml"

  | 2 ->
# 679 "src/reason-parser/reason_declarative_lexer.mll"
    ( set_lexeme_length lexbuf 2;
      let loc = Location.curr lexbuf in
      let raw_buffer, _ = get_scratch_buffers state in
      ignore (comment raw_buffer loc loc lexbuf : bool);
      lexbuf.Lexing.lex_start_p <- loc.Location.loc_start;
      let loc_end = lexbuf.Lexing.lex_curr_p in
      COMMENT (flush_buffer raw_buffer,
               {loc with Location.loc_end})
    )
# 3687 "src/reason-parser/reason_declarative_lexer.ml"

  | 3 ->
# 689 "src/reason-parser/reason_declarative_lexer.mll"
    ( let loc = Location.curr lexbuf in
      let raw_buffer, _ = get_scratch_buffers state in
      ignore (comment raw_buffer loc loc lexbuf : bool);
      lexbuf.Lexing.lex_start_p <- loc.Location.loc_start;
      DOCSTRING (flush_buffer raw_buffer)
    )
# 3697 "src/reason-parser/reason_declarative_lexer.ml"

  | 4 ->
# 696 "src/reason-parser/reason_declarative_lexer.mll"
    ( DOCSTRING "" )
# 3702 "src/reason-parser/reason_declarative_lexer.ml"

  | 5 ->
# 698 "src/reason-parser/reason_declarative_lexer.mll"
    ( let loc = Location.curr lexbuf in
      Location.prerr_warning loc Warnings.Comment_start;
      let raw_buffer, _ = get_scratch_buffers state in
      ignore (comment raw_buffer loc loc lexbuf : bool);
      let loc_end = lexbuf.Lexing.lex_curr_p in
      COMMENT (flush_buffer raw_buffer,
               {loc with Location.loc_end})
    )
# 3714 "src/reason-parser/reason_declarative_lexer.ml"

  | 6 ->
# 707 "src/reason-parser/reason_declarative_lexer.mll"
    ( let loc = Location.curr lexbuf in
      Location.prerr_warning loc Warnings.Comment_not_end;
      set_lexeme_length lexbuf 1;
      STAR
    )
# 3723 "src/reason-parser/reason_declarative_lexer.ml"

  | 7 ->
# 712 "src/reason-parser/reason_declarative_lexer.mll"
      ( assert false )
# 3728 "src/reason-parser/reason_declarative_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_enter_comment_rec state lexbuf __ocaml_lex_state

and comment buffer firstloc nestedloc lexbuf =
   __ocaml_lex_comment_rec buffer firstloc nestedloc lexbuf 212
and __ocaml_lex_comment_rec buffer firstloc nestedloc lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 721 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      if comment buffer firstloc (Location.curr lexbuf) lexbuf then (
        store_lexeme buffer lexbuf;
        comment buffer firstloc nestedloc lexbuf
      )
      else
        false
    )
# 3747 "src/reason-parser/reason_declarative_lexer.ml"

  | 1 ->
# 730 "src/reason-parser/reason_declarative_lexer.mll"
    ( true )
# 3752 "src/reason-parser/reason_declarative_lexer.ml"

  | 2 ->
# 732 "src/reason-parser/reason_declarative_lexer.mll"
    ( Buffer.add_char buffer '"';
      let string_start = Location.curr lexbuf in
      let terminated_string = string buffer None lexbuf in
      Buffer.add_char buffer '"';
      if terminated_string then
        comment buffer firstloc nestedloc lexbuf
      else (
        raise_error nestedloc
          (Unterminated_string_in_comment (firstloc, string_start));
        false
      )
    )
# 3768 "src/reason-parser/reason_declarative_lexer.ml"

  | 3 ->
let
# 744 "src/reason-parser/reason_declarative_lexer.mll"
                       delim
# 3774 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 745 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      let stringloc = Location.curr lexbuf in
      let terminated_string = quoted_string buffer delim lexbuf in
      Buffer.add_char buffer '|';
      Buffer.add_string buffer delim;
      Buffer.add_char buffer '}';
      if terminated_string then
        comment buffer firstloc nestedloc lexbuf
      else (
        raise_error nestedloc
          (Unterminated_string_in_comment (firstloc, stringloc));
        false
      )
    )
# 3791 "src/reason-parser/reason_declarative_lexer.ml"

  | 4 ->
# 760 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      comment buffer firstloc nestedloc lexbuf
    )
# 3798 "src/reason-parser/reason_declarative_lexer.ml"

  | 5 ->
# 764 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      update_loc lexbuf None 1 false 1;
      comment buffer firstloc nestedloc lexbuf
    )
# 3806 "src/reason-parser/reason_declarative_lexer.ml"

  | 6 ->
# 772 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      comment buffer firstloc nestedloc lexbuf
    )
# 3813 "src/reason-parser/reason_declarative_lexer.ml"

  | 7 ->
# 776 "src/reason-parser/reason_declarative_lexer.mll"
    ( raise_error nestedloc (Unterminated_comment firstloc);
      false
    )
# 3820 "src/reason-parser/reason_declarative_lexer.ml"

  | 8 ->
# 780 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      update_loc lexbuf None 1 false 0;
      comment buffer firstloc nestedloc lexbuf
    )
# 3828 "src/reason-parser/reason_declarative_lexer.ml"

  | 9 ->
# 785 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      comment buffer firstloc nestedloc lexbuf
    )
# 3835 "src/reason-parser/reason_declarative_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_rec buffer firstloc nestedloc lexbuf __ocaml_lex_state

and string rawbuf txtbuf lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1); __ocaml_lex_string_rec rawbuf txtbuf lexbuf 241
and __ocaml_lex_string_rec rawbuf txtbuf lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 800 "src/reason-parser/reason_declarative_lexer.mll"
    ( true )
# 3847 "src/reason-parser/reason_declarative_lexer.ml"

  | 1 ->
let
# 801 "src/reason-parser/reason_declarative_lexer.mll"
                                  space
# 3853 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 802 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme rawbuf lexbuf;
      update_loc lexbuf None 1 false (String.length space);
      string rawbuf txtbuf lexbuf
    )
# 3860 "src/reason-parser/reason_declarative_lexer.ml"

  | 2 ->
let
# 806 "src/reason-parser/reason_declarative_lexer.mll"
                                                 c
# 3866 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1) in
# 807 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme rawbuf lexbuf;
      begin match txtbuf with
      | None -> ()
      | Some buf -> Buffer.add_char buf (char_for_backslash c);
      end;
      string rawbuf txtbuf lexbuf
    )
# 3876 "src/reason-parser/reason_declarative_lexer.ml"

  | 3 ->
# 815 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme rawbuf lexbuf;
      begin match txtbuf with
      | None -> ()
      | Some buf -> Buffer.add_char buf (char_for_decimal_code lexbuf 1);
      end;
      string rawbuf txtbuf lexbuf
    )
# 3887 "src/reason-parser/reason_declarative_lexer.ml"

  | 4 ->
# 823 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme rawbuf lexbuf;
      begin match txtbuf with
      | None -> ()
      | Some buf -> Buffer.add_char buf (char_for_hexadecimal_code lexbuf 2);
      end;
      string rawbuf txtbuf lexbuf
    )
# 3898 "src/reason-parser/reason_declarative_lexer.ml"

  | 5 ->
# 831 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme rawbuf lexbuf;
      begin match txtbuf with
      | None -> ()
      | Some buf ->
         store_lexeme buf lexbuf;
         (*  FIXME: Warnings should probably go in Reason_errors
            Should be an error, but we are very lax.
              raise (Error (Illegal_escape (Lexing.lexeme lexbuf),
                        Location.curr lexbuf))
           FIXME Using Location relies too much on compiler internals
          *)
         Location.prerr_warning (Location.curr lexbuf)
           Warnings.Illegal_backslash;
      end;
      string rawbuf txtbuf lexbuf
    )
# 3918 "src/reason-parser/reason_declarative_lexer.ml"

  | 6 ->
# 848 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme rawbuf lexbuf;
      begin match txtbuf with
      | None -> ()
      | Some buf ->
        store_lexeme buf lexbuf;
        Location.prerr_warning (Location.curr lexbuf)
          Warnings.Eol_in_string
      end;
      update_loc lexbuf None 1 false 0;
      string rawbuf txtbuf lexbuf
    )
# 3933 "src/reason-parser/reason_declarative_lexer.ml"

  | 7 ->
# 860 "src/reason-parser/reason_declarative_lexer.mll"
    ( false )
# 3938 "src/reason-parser/reason_declarative_lexer.ml"

  | 8 ->
# 862 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme rawbuf lexbuf;
      begin match txtbuf with
      | None -> ()
      | Some buf -> Buffer.add_char buf (Lexing.lexeme_char lexbuf 0);
      end;
      string rawbuf txtbuf lexbuf
    )
# 3949 "src/reason-parser/reason_declarative_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_rec rawbuf txtbuf lexbuf __ocaml_lex_state

and quoted_string buffer delim lexbuf =
   __ocaml_lex_quoted_string_rec buffer delim lexbuf 260
and __ocaml_lex_quoted_string_rec buffer delim lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 881 "src/reason-parser/reason_declarative_lexer.mll"
    ( store_lexeme buffer lexbuf;
      update_loc lexbuf None 1 false 0;
      quoted_string buffer delim lexbuf
    )
# 3964 "src/reason-parser/reason_declarative_lexer.ml"

  | 1 ->
# 886 "src/reason-parser/reason_declarative_lexer.mll"
    ( false )
# 3969 "src/reason-parser/reason_declarative_lexer.ml"

  | 2 ->
let
# 887 "src/reason-parser/reason_declarative_lexer.mll"
                       edelim
# 3975 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 888 "src/reason-parser/reason_declarative_lexer.mll"
    ( if delim = edelim then
        true
      else (
        store_lexeme buffer lexbuf;
        quoted_string buffer delim lexbuf
      )
    )
# 3985 "src/reason-parser/reason_declarative_lexer.ml"

  | 3 ->
let
# 895 "src/reason-parser/reason_declarative_lexer.mll"
         c
# 3991 "src/reason-parser/reason_declarative_lexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 896 "src/reason-parser/reason_declarative_lexer.mll"
    ( Buffer.add_char buffer c;
      quoted_string buffer delim lexbuf
    )
# 3997 "src/reason-parser/reason_declarative_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_quoted_string_rec buffer delim lexbuf __ocaml_lex_state

and skip_sharp_bang lexbuf =
   __ocaml_lex_skip_sharp_bang_rec lexbuf 269
and __ocaml_lex_skip_sharp_bang_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 902 "src/reason-parser/reason_declarative_lexer.mll"
       ( update_loc lexbuf None 3 false 0 )
# 4009 "src/reason-parser/reason_declarative_lexer.ml"

  | 1 ->
# 904 "src/reason-parser/reason_declarative_lexer.mll"
       ( update_loc lexbuf None 1 false 0 )
# 4014 "src/reason-parser/reason_declarative_lexer.ml"

  | 2 ->
# 905 "src/reason-parser/reason_declarative_lexer.mll"
       ( () )
# 4019 "src/reason-parser/reason_declarative_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_skip_sharp_bang_rec lexbuf __ocaml_lex_state

;;

